/*------------------------------------------------------------------------------*
 * Architecture & Implementation of DBMS                                        *
 *------------------------------------------------------------------------------*
 * Copyright 2020 Databases and Information Systems Group TU Dortmund           *
 * Visit us at                                                                  *
 *             http://dbis.cs.tu-dortmund.de/cms/en/home/                       *
 *                                                                              *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS      *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,  *
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL      *
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR         *
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,        *
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR        *
 * OTHER DEALINGS IN THE SOFTWARE.                                              *
 *                                                                              *
 * Authors (alphabetical list):                                                 *
 *          Maximilian Berens   <maximilian.berens@tu-dortmund.de>              *
 *          Roland Kühn         <roland.kuehn@cs.tu-dortmund.de>                *
 *          Jan Mühlig          <jan.muehlig@tu-dortmund.de>                    *
 *------------------------------------------------------------------------------*
 */

#include <argparse/argparse.hpp>
#include <config.h>
#include <database.h>
#include <io/file_executor.h>
#include <io/printing_executor.h>
#include <io/user_console.h>
#include <parser/query_parser.h>
#include <string>
#include <util/ini_parser.h>

int main(int arg_count, char **args)
{
    // Read configuration file.
    beedb::util::IniParser ini_parser{"beedb.ini"};
    if (ini_parser.empty())
    {
        std::cout << "[Warning] Missing configuration file beedb.ini" << std::endl;
    }

    const auto buffer_frames = ini_parser.get<std::uint32_t>("buffer manager", "frames", 256u);
    const auto scan_page_limit = ini_parser.get<std::uint32_t>("scan", "page-limit", 64u);
    const auto enable_index_scan = ini_parser.get<bool>("scan", "enable-index-scan", false);
    const auto buffer_replacement_strategy = ini_parser.get<std::string>("buffer manager", "strategy", "LRU");
    const auto lru_k = ini_parser.get<std::uint32_t>("buffer manager", "k", 2u);
    const auto enable_hash_join = ini_parser.get<bool>("join", "enable-hash-join", false);
    const auto print_statistics = ini_parser.get<bool>("execution", "print-statistics", false);

    // Parse command line arguments
    argparse::ArgumentParser argument_parser("beedb");
    argument_parser.add_argument("db-file")
        .help("File the database is stored in. Default: bee.db")
        .default_value(std::string("bee.db"));
    argument_parser.add_argument("-l", "--load").help("Load SQL file into database.").default_value(std::string(""));
    argument_parser.add_argument("-q", "--query").help("Execute Query.").default_value(std::string(""));
    argument_parser.add_argument("-c", "--console")
        .help("Run console after loading file or executing query.")
        .implicit_value(true)
        .default_value(false);
    argument_parser.add_argument("--buffer-manager-frames")
        .help("Number of frames within the frame buffer.")
        .default_value(buffer_frames)
        .action([](const std::string &value) { return std::uint32_t(std::stoi(value)); });
    argument_parser.add_argument("--scan-page-limit")
        .help("Number of pages the SCAN operator can pin at a time.")
        .default_value(scan_page_limit)
        .action([](const std::string &value) { return std::uint32_t(std::stoi(value)); });
    argument_parser.add_argument("--enable-index-scan")
        .help("Enable index scan and use whenever possible.")
        .implicit_value(true)
        .default_value(enable_index_scan);
    argument_parser.add_argument("--enable-hash-join")
        .help("Enable hash join and use whenever possible.")
        .implicit_value(true)
        .default_value(enable_hash_join);
    argument_parser.add_argument("--stats")
        .help("Print all execution statistics")
        .implicit_value(true)
        .default_value(print_statistics);

    try
    {
        argument_parser.parse_args(arg_count, args);
    }
    catch (std::runtime_error &)
    {
        argument_parser.print_help();
        return 1;
    }

    std::regex lfu_regex("lfu", std::regex::icase);
    std::regex lru_k_regex("lru-k", std::regex::icase);
    std::smatch match;
    beedb::Config::BufferReplacementStrategy strategy = beedb::Config::LRU;
    if (std::regex_match(buffer_replacement_strategy, match, lfu_regex))
    {
        strategy = beedb::Config::LFU;
    }
    else if (std::regex_match(buffer_replacement_strategy, match, lru_k_regex))
    {
        strategy = beedb::Config::LRU_K;
    }

    beedb::Config config{};

    config.set(beedb::Config::k_BufferFrames, argument_parser.get<std::uint32_t>("--buffer-manager-frames"),
               beedb::Config::ConfigMapValue::immutable);
    config.set(beedb::Config::k_BufferReplacementStrategy, strategy, beedb::Config::ConfigMapValue::immutable);
    config.set(beedb::Config::k_LRU_K, lru_k, beedb::Config::ConfigMapValue::immutable);
    config.set(beedb::Config::k_ScanPageLimit, argument_parser.get<std::uint32_t>("--scan-page-limit"),
               beedb::Config::ConfigMapValue::immutable);
    config.set(beedb::Config::k_OptimizationEnableIndexScan, argument_parser.get<bool>("--enable-index-scan"));
    config.set(beedb::Config::k_OptimizationEnableHashJoin, argument_parser.get<bool>("--enable-hash-join"));
    config.set(beedb::Config::k_OptimizationDisableOptimization,
               true); // true (at first), to disable optimization during boot
    config.set(beedb::Config::k_CheckFinalPlan, false);
    config.set(beedb::Config::k_PrintExecutionStatistics, argument_parser.get<bool>("--stats"));

    const auto database_file_name = argument_parser.get<std::string>("db-file");
    const auto sql_file = argument_parser.get<std::string>("-l");
    const auto query = argument_parser.get<std::string>("-q");
    const auto is_console = argument_parser.get<bool>("-c");

    beedb::Database database(config, database_file_name);
    database.boot();

    // re-enable optimization for user-queries, after boot code ran:
    config.set(beedb::Config::k_OptimizationDisableOptimization, false);

    // re-enable checks on all plans, when in debug mode:
    assert(config.set(beedb::Config::k_CheckFinalPlan, true));

    if (sql_file.empty() == false)
    {
        beedb::io::FileExecutor file_importer(database);
        file_importer.import_file(sql_file);
    }

    if (query.empty() == false)
    {
        beedb::io::PrintingExecutor printing_executor(database);
        printing_executor.execute(beedb::io::Query{query});
    }

    if ((sql_file.empty() == false || query.empty() == false) && is_console == false)
    {
        return 0;
    }

    beedb::io::UserConsole user_console(database);
    user_console.wait_for_input();

    return 0;
}