/*------------------------------------------------------------------------------*
 * Architecture & Implementation of DBMS                                        *
 *------------------------------------------------------------------------------*
 * Copyright 2020 Databases and Information Systems Group TU Dortmund           *
 * Visit us at                                                                  *
 *             http://dbis.cs.tu-dortmund.de/cms/en/home/                       *
 *                                                                              *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS      *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,  *
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL      *
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR         *
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,        *
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR        *
 * OTHER DEALINGS IN THE SOFTWARE.                                              *
 *                                                                              *
 * Authors (alphabetical list):                                                 *
 *          Maximilian Berens   <maximilian.berens@tu-dortmund.de>              *
 *          Roland Kühn         <roland.kuehn@cs.tu-dortmund.de>                *
 *          Jan Mühlig          <jan.muehlig@tu-dortmund.de>                    *
 *------------------------------------------------------------------------------*
 */

#include <cassert>
#include <config.h>
#include <database.h>
#include <disk/buffer_replacement_strategy.h>
#include <disk/lfu_strategy.h>
#include <disk/lru_k_strategy.h>
#include <disk/lru_strategy.h>
#include <index/index_factory.h>
#include <io/executor.h>
#include <plan/physical/builder.h>
#include <sstream>
#include <table/column.h>

using namespace beedb;

Database::Database(Config &config, const std::string &file_name)
    : _config(config), _storage_manager(file_name), _buffer_manager(config[Config::k_BufferFrames], _storage_manager),
      _table_disk_manager(_buffer_manager)
{
    // Initialize BufferManagerStrategy.
    std::unique_ptr<disk::BufferReplacementStrategy> replacement_strategy{};
    if (config[Config::k_BufferReplacementStrategy] == Config::LFU)
    {
        replacement_strategy.reset(new disk::LFUStrategy());
    }
    else if (config[Config::k_BufferReplacementStrategy] == Config::LRU_K)
    {
        replacement_strategy.reset(new disk::LRUKStrategy(config[Config::k_LRU_K]));
    }
    else
    {
        replacement_strategy.reset(new disk::LRUStrategy());
    }

    this->_buffer_manager.replacement_strategy(std::move(replacement_strategy));
}

Database::~Database()
{
    // Update statistics
    for (auto [_, table] : this->_tables)
    {
        if (table->is_virtual() == false)
        {
            this->persist_table_statistics(table, this->_statistics.table_statistics().cardinality(*table));
        }
    }

    // Delete tables AFTER all statistics are persisted.
    // Otherwise it is possible to delete the statistics table
    // before all updates are done.
    for (auto [_, table] : this->_tables)
    {
        delete table;
    }
}

void Database::boot()
{
    // Initialize tables with fixed schema and allocate pages for the data,
    // if the file is empty.
    this->initialize_database(this->_storage_manager.count_pages() == 0);

    // Read all tables, columns and indices and build tables from the results.
    auto tables_executor = io::Executor{*this};
    tables_executor.execute(io::Query{"select * from system_tables"}, [&](const table::Tuple &table_tuple) {
        const auto table_id = table_tuple.get(0).get<std::int32_t>();
        const auto table_name = table_tuple.get(1).get<std::string>();
        const auto table_page = table_tuple.get(2).get<std::int32_t>();

        auto schema = table::Schema{table_name};
        auto columns_executor = io::Executor{*this};
        columns_executor.execute(
            io::Query{"select * from system_columns where table_id = " + std::to_string(table_id)},
            [&](const table::Tuple &column_tuple) {
                const auto column_id = column_tuple.get(0).get<std::int32_t>();
                const auto column_type_id = column_tuple.get(2).get<std::int32_t>();
                const auto column_length = column_tuple.get(3).get<std::int32_t>();
                const auto column_name = column_tuple.get(4).get<std::string>();
                const auto column_is_nullable = column_tuple.get(5).get<std::int32_t>();
                const auto type =
                    table::Type(static_cast<table::Type::Id>(column_type_id), std::uint16_t(column_length));

                std::vector<std::shared_ptr<index::IndexInterface>> indices;
                auto indices_executor = io::Executor{*this};
                indices_executor.execute(
                    io::Query{"select * from system_indices where column_id = " + std::to_string(column_id)},
                    [&](const table::Tuple &index_tuple) {
                        const auto index_id = index_tuple.get(0).get<std::int32_t>();
                        const auto index_type_id = index_tuple.get(2).get<std::int32_t>();
                        const auto index_name = index_tuple.get(3).get<std::string>();
                        const auto is_unique = index_tuple.get(4).get<std::int32_t>();

                        indices.push_back(index::IndexFactory::new_index(
                            index_name, static_cast<index::Type>(index_type_id), static_cast<bool>(is_unique)));
                        this->_next_index_id = std::max(index_id + 1u, this->_next_index_id);
                    });

                schema.add({column_id, type, static_cast<bool>(column_is_nullable), std::move(indices)},
                           {column_name, table_name});
                this->_next_column_id = std::max(column_id + 1u, this->_next_column_id);
            });

        this->_tables[table_name] = new table::Table(table_id, table_page, std::move(schema));

        auto statistics_executor = io::Executor{*this};
        statistics_executor.execute(
            {"select * from system_table_statistics where table_id = " + std::to_string(table_id)},
            [this, table_id](const table::Tuple &tuple) {
                const auto cardinality = tuple.get(1).get<std::int64_t>();
                this->_statistics.table_statistics().cardinality(table_id, cardinality);
            });

        this->_next_table_id = std::max(table_id + 1u, this->_next_table_id);
    });

    // Build and fill all indices.
    io::Executor build_index_executor{*this};
    for (auto &[table_name, table] : this->_tables)
    {
        for (auto i = 0u; i < table->schema().size(); i++)
        {
            const auto &column = table->schema().column(i);
            const auto &attribute = table->schema().attribute(i);
            for (auto index : column.indices())
            {
                auto plan = plan::physical::Builder::build_index_plan(*this, table_name, attribute.name, index->name());
                build_index_executor.execute(plan);
            }
        }
    }
}

void Database::initialize_database(const bool create_schema)
{
    if (create_schema)
    {
        // Allocate page for tables.
        auto tables_page = this->_buffer_manager.allocate();
        assert(tables_page->id() == 0);
        this->_buffer_manager.unpin(tables_page, false);

        // Allocate page for columns.
        auto columns_page = this->_buffer_manager.allocate();
        assert(columns_page->id() == 1);
        this->_buffer_manager.unpin(columns_page, false);

        // Allocate page for indices.
        auto indices_page = this->_buffer_manager.allocate();
        assert(indices_page->id() == 2);
        this->_buffer_manager.unpin(indices_page, false);

        // Allocate page for indices.
        auto table_statistics_page = this->_buffer_manager.allocate();
        assert(table_statistics_page->id() == 3);
        this->_buffer_manager.unpin(table_statistics_page, false);
    }

    std::string tables_name("system_tables");
    table::Schema tables_schema{tables_name};
    tables_schema.add({table::Type::INT, false}, {"id", tables_schema.table_name()});
    tables_schema.add({{table::Type::CHAR, 48}, false}, {"name", tables_schema.table_name()});
    tables_schema.add({table::Type::INT, false}, {"page", tables_schema.table_name()});

    auto tables_table = new table::Table(-1, 0, std::move(tables_schema));

    std::string columns_name("system_columns");
    table::Schema columns_schema{columns_name};
    columns_schema.add({table::Type::INT, false}, {"id", columns_schema.table_name()});
    columns_schema.add({table::Type::INT, false}, {"table_id", columns_schema.table_name()});
    columns_schema.add({table::Type::INT, false}, {"type_id", columns_schema.table_name()});
    columns_schema.add({table::Type::INT, false}, {"length", columns_schema.table_name()});
    columns_schema.add({{table::Type::CHAR, 48}, false}, {"name", columns_schema.table_name()});
    columns_schema.add({table::Type::INT, false}, {"is_nullable", columns_schema.table_name()});
    columns_schema.add({table::Type::INT, false}, {"is_unique", columns_schema.table_name()});
    columns_schema.add({table::Type::INT, false}, {"is_primary_key", columns_schema.table_name()});
    auto columns_table = new table::Table(-1, 1, std::move(columns_schema));

    std::string indices_name("system_indices");
    table::Schema indices_schema{indices_name};
    indices_schema.add({table::Type::INT, false}, {"id", indices_schema.table_name()});
    indices_schema.add({table::Type::INT, false}, {"column_id", indices_schema.table_name()});
    indices_schema.add({table::Type::INT, false}, {"type_id", indices_schema.table_name()});
    indices_schema.add({{table::Type::CHAR, 48}, false}, {"name", indices_schema.table_name()});
    indices_schema.add({table::Type::INT, false}, {"is_unique", indices_schema.table_name()});
    auto indices_table = new table::Table(-1, 2, std::move(indices_schema));

    std::string table_statistics_name("system_table_statistics");
    table::Schema table_statistics_schema{table_statistics_name};
    table_statistics_schema.add({table::Type::INT, false}, {"table_id", table_statistics_schema.table_name()});
    table_statistics_schema.add({table::Type::LONG, false}, {"cardinality", table_statistics_schema.table_name()});
    auto table_statistics_table = new table::Table(-1, 3, std::move(table_statistics_schema));

    this->_tables[tables_table->name()] = tables_table;
    this->_tables[columns_table->name()] = columns_table;
    this->_tables[indices_table->name()] = indices_table;
    this->_tables[table_statistics_table->name()] = table_statistics_table;
}

void Database::create_table(const table::Schema &schema)
{
    auto tables_table = this->_tables["system_tables"];
    auto columns_table = this->_tables["system_columns"];
    auto table_statistics_table = this->_tables["system_table_statistics"];

    // Persist table.
    table::Tuple table_tuple(tables_table->schema(), tables_table->schema().row_size());
    auto table_id = std::int32_t(this->_next_table_id++);
    auto name = schema.table_name();
    auto page = this->_buffer_manager.allocate();
    auto page_id = std::int32_t(page->id());
    this->_buffer_manager.unpin(page, false);
    table_tuple.set(0, table_id);
    table_tuple.set(1, name);
    table_tuple.set(2, page_id);
    this->_table_disk_manager.add_row(*tables_table, std::move(table_tuple));

    // Persist columns
    for (auto i = 0u; i < schema.size(); i++)
    {
        const auto &column = schema.column(i);
        const auto &attribute = schema.attribute(i);
        table::Tuple column_tuple(columns_table->schema(), columns_table->schema().row_size());
        auto column_id = std::int32_t(this->_next_column_id++);
        auto column_name = attribute.name;
        auto type_id = std::int32_t(static_cast<table::Type::Id>(column.type()));
        auto length = std::int32_t(column.type().dynamic_length());
        auto is_nullable = std::int32_t(column.is_nullable());
        auto is_unique = std::int32_t(0);
        auto is_primary_key = std::int32_t(0);
        column_tuple.set(0, column_id);
        column_tuple.set(1, table_id);
        column_tuple.set(2, type_id);
        column_tuple.set(3, length);
        column_tuple.set(4, column_name);
        column_tuple.set(5, is_nullable);
        column_tuple.set(6, is_unique);
        column_tuple.set(7, is_primary_key);
        this->_table_disk_manager.add_row(*columns_table, std::move(column_tuple));
    }

    table::Tuple table_statistics_tuple(table_statistics_table->schema(), table_statistics_table->schema().row_size());
    auto cardinality = std::int64_t{0};
    table_statistics_tuple.set(0, table_id);
    table_statistics_tuple.set(1, cardinality);
    this->_table_disk_manager.add_row(*table_statistics_table, std::move(table_statistics_tuple));

    this->_tables[schema.table_name()] = new table::Table(table_id, page_id, schema);
}

void Database::create_index(const table::Column &column, const index::Type type, const std::string &name,
                            const bool is_unique)
{
    auto *indices_table = this->_tables["system_indices"];
    table::Tuple tuple(indices_table->schema(), indices_table->schema().row_size());
    auto index_id = std::int32_t(this->_next_index_id++);
    auto column_id = std::int32_t(column.id());
    auto type_id = std::int32_t(type);
    auto index_name = name;
    auto unique = std::int32_t(is_unique);
    tuple.set(0, index_id);
    tuple.set(1, column_id);
    tuple.set(2, type_id);
    tuple.set(3, index_name);
    tuple.set(4, unique);
    this->_table_disk_manager.add_row(*indices_table, std::move(tuple));
}

void Database::persist_table_statistics(beedb::table::Table *table, const std::uint64_t cardinality)
{
    io::Executor executor(*this);
    executor.execute({"update system_table_statistics set cardinality = " + std::to_string(std::int64_t(cardinality)) +
                      " where table_id = " + std::to_string(table->id()) + ";"});
}