/*------------------------------------------------------------------------------*
 * Architecture & Implementation of DBMS                                        *
 *------------------------------------------------------------------------------*
 * Copyright 2020 Databases and Information Systems Group TU Dortmund           *
 * Visit us at                                                                  *
 *             http://dbis.cs.tu-dortmund.de/cms/en/home/                       *
 *                                                                              *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS      *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,  *
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL      *
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR         *
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,        *
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR        *
 * OTHER DEALINGS IN THE SOFTWARE.                                              *
 *                                                                              *
 * Authors (alphabetical list):                                                 *
 *          Maximilian Berens   <maximilian.berens@tu-dortmund.de>              *
 *          Roland Kühn         <roland.kuehn@cs.tu-dortmund.de>                *
 *          Jan Mühlig          <jan.muehlig@tu-dortmund.de>                    *
 *------------------------------------------------------------------------------*
 */

#include <cassert>
#include <config.h>
#include <disk/storage_manager.h>
#include <exception/disk_exception.h>

using namespace beedb::disk;

StorageManager::StorageManager(const std::string &file_name)
{
    this->_storage_file.open(file_name, std::ios::in | std::ios::out | std::ios::binary);
    if (this->_storage_file.is_open() == false)
    {
        // Create file if not exists. Non-existing files can not be opened with in | out only -> add trunc.
        this->_storage_file.open(file_name, std::ios::in | std::ios::out | std::ios::binary | std::ios::trunc);
    }

    if (this->_storage_file.is_open() == false)
    {
        throw exception::CanNotOpenStorageFile();
    }
    this->_storage_file.seekg(0, std::ios::end);

    assert(this->_storage_file.tellg() % Config::page_size == 0);
    this->_count_pages = this->_storage_file.tellg() / Config::page_size;

    assert(sizeof(Page) == Config::page_size);
}

StorageManager::~StorageManager()
{
    this->_storage_file.close();
}

Page StorageManager::read(const Page::page_id page_id)
{
    Page page(page_id);
    this->_storage_file.seekg(page.id() * sizeof(Page), std::ios::beg);
    this->_storage_file.read(reinterpret_cast<char *>(&page), sizeof(Page));
    return page;
}

void StorageManager::write(Page &page)
{
    this->_storage_file.seekp(page.id() * sizeof(Page), std::ios::beg);
    this->_storage_file.write(reinterpret_cast<char *>(&page), sizeof(Page));
}

Page::page_id StorageManager::allocate()
{
    Page page(this->_count_pages++);
    this->write(page);
    return page.id();
}