/*------------------------------------------------------------------------------*
 * Architecture & Implementation of DBMS                                        *
 *------------------------------------------------------------------------------*
 * Copyright 2020 Databases and Information Systems Group TU Dortmund           *
 * Visit us at                                                                  *
 *             http://dbis.cs.tu-dortmund.de/cms/en/home/                       *
 *                                                                              *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS      *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,  *
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL      *
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR         *
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,        *
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR        *
 * OTHER DEALINGS IN THE SOFTWARE.                                              *
 *                                                                              *
 * Authors (alphabetical list):                                                 *
 *          Maximilian Berens   <maximilian.berens@tu-dortmund.de>              *
 *          Roland Kühn         <roland.kuehn@cs.tu-dortmund.de>                *
 *          Jan Mühlig          <jan.muehlig@tu-dortmund.de>                    *
 *------------------------------------------------------------------------------*
 */

#pragma once
#include <config.h>
#include <cstdint>
#include <disk/buffer_manager.h>
#include <disk/lru_strategy.h>
#include <disk/page.h>
#include <disk/storage_manager.h>
#include <functional>
#include <index/type.h>
#include <statistic/system_statistics.h>
#include <string>
#include <table/table.h>
#include <table/table_disk_manager.h>
#include <table/tuple.h>
#include <unordered_map>
#include <utility>

namespace beedb
{
class Database
{
  public:
    Database(Config &config, const std::string &file_name);
    ~Database();

    /**
     * Boots the database management system.
     * During the boot, all persisted tables, their schemas and indices
     * will be loaded to memory.
     * All indices will be filled with data from disk.
     */
    void boot();

    /**
     * @return Instance of the TableDiskManager.
     */
    inline table::TableDiskManager &table_disk_manager()
    {
        return _table_disk_manager;
    }

    /**
     * @return Instance of the BufferManager.
     */
    inline disk::BufferManager &buffer_manager()
    {
        return _buffer_manager;
    }

    /**
     * @return Immutable instance of the config.
     */
    inline const Config &config() const
    {
        return _config;
    }

    /**
     * @return Mutable instance of the config.
     */
    inline Config &config()
    {
        return _config;
    }

    inline statistic::SystemStatistics &system_statistics()
    {
        return _statistics;
    }

    /**
     * Checks whether a table exists.
     *
     * @param name Name of the table.
     * @return True, if the table exists.
     */
    inline bool table_exists(const std::string &name)
    {
        return _tables.find(name) != _tables.end();
    }

    /**
     * Returns a pointer to the requested table.
     *
     * @param name Name of the requested table.
     * @return Pointer to the table.
     */
    table::Table *table(const std::string &name)
    {
        if (table_exists(name))
        {
            return _tables[name];
        }

        return nullptr;
    }

    table::Table *operator[](const std::string &table_name)
    {
        return table(table_name);
    }

    const table::Table &operator[](const std::string &table_name) const
    {
        // unchecked, const access to tables
        return *(_tables.at(table_name));
    }

    /**
     * Creates a table with a given schema.
     * The table will be persisted and available after creation.
     *
     * @param schema Schema for the table.
     */
    void create_table(const table::Schema &schema);

    /**
     * Creates an index for a specific column.
     * The index will be persisted, filled, and available after creation.
     *
     * @param column Column to be indexed.
     * @param type Type of the index.
     * @param name Name of the index.
     * @param is_unique True, when the index is a unique index.
     */
    void create_index(const table::Column &column, const index::Type type, const std::string &name,
                      const bool is_unique);

  private:
    Config &_config;
    disk::StorageManager _storage_manager;
    disk::BufferManager _buffer_manager;
    table::TableDiskManager _table_disk_manager;

    std::unordered_map<std::string, table::Table *> _tables;

    std::uint32_t _next_table_id = 1;
    std::uint32_t _next_column_id = 1;
    std::uint32_t _next_index_id = 1;

    statistic::SystemStatistics _statistics;

    /**
     * Initializes the database. When the database is empty,
     * we will create a new database schema containing all meta tables.
     *
     * @param create_schema True, when a database schema should be created.
     */
    void initialize_database(const bool create_schema);

    /**
     * Persists the table statistics.
     *
     * @param table Table
     * @param cardinality Cardinality
     */
    void persist_table_statistics(table::Table *table, const std::uint64_t cardinality);
};
} // namespace beedb