/*------------------------------------------------------------------------------*
 * Architecture & Implementation of DBMS                                        *
 *------------------------------------------------------------------------------*
 * Copyright 2020 Databases and Information Systems Group TU Dortmund           *
 * Visit us at                                                                  *
 *             http://dbis.cs.tu-dortmund.de/cms/en/home/                       *
 *                                                                              *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS      *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,  *
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL      *
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR         *
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,        *
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR        *
 * OTHER DEALINGS IN THE SOFTWARE.                                              *
 *                                                                              *
 * Authors (alphabetical list):                                                 *
 *          Maximilian Berens   <maximilian.berens@tu-dortmund.de>              *
 *          Roland Kühn         <roland.kuehn@cs.tu-dortmund.de>                *
 *          Jan Mühlig          <jan.muehlig@tu-dortmund.de>                    *
 *------------------------------------------------------------------------------*
 */

#pragma once
#include "buffer_replacement_strategy.h"
#include "frame.h"
#include "page.h"
#include "storage_manager.h"
#include <cstdint>
#include <memory>
#include <vector>

namespace beedb::disk
{
/**
 * The BufferManager buffers pages stored on the disk in memory.
 * Since the system has not infinite memory, the number of buffered
 * pages (named "frames") is limited.
 *
 * Every access to a page on the disk is done by pinning the page
 * through the BufferManager. When the page is not needed any more
 * (e.g. all tuples are scanned), the page can be unpinned by the
 * BufferManager.
 */
class BufferManager
{
  public:
    BufferManager(const std::size_t count_frames, StorageManager &space_manager,
                  std::unique_ptr<BufferReplacementStrategy> replacement_strategy = nullptr);
    ~BufferManager();

    /**
     * Loads the page from disk into memory and returns a pointer
     * to the page. When the page is still buffered, the page will
     * not be loaded twice, but guaranteed to stay in memory until
     * it is unpinned.
     *
     * @param page_id Id of the page.
     * @return Pointer to the page, that allows accessing the data.
     */
    Page *pin(const Page::page_id page_id);

    /**
     * Notifies the BufferManager that the page is not needed anymore.
     * In case no one needs the page, the frame can be used for other
     * pages buffered from disk in memory.
     *
     * @param page_id   Id of the page.
     * @param is_dirty  True, when the content of the page was modified.
     */
    void unpin(const Page::page_id page_id, const bool is_dirty);

    /**
     * Notifies the BufferManager that the page is not needed anymore.
     * In case no one needs the page, the frame can be used for other
     * pages buffered from disk in memory.
     *
     * @param page_id   Id of the page.
     * @param is_dirty  True, when the content of the page was modified.
     */
    inline void unpin(Page *page, const bool is_dirty)
    {
        unpin(page->id(), is_dirty);
    }

    /**
     * Allocates a new page on the disk and loads the page to memory.
     *
     * @return Pointer to the pinned(!) page.
     */
    Page *allocate();

    /**
     * Set the replacement strategy which picks frames to be replaced,
     * when all frames are occupied, but a new page is requested to
     * be loaded from disk to memory.
     * @param replacement_strategy
     */
    inline void replacement_strategy(std::unique_ptr<BufferReplacementStrategy> replacement_strategy)
    {
        _replacement_strategy = std::move(replacement_strategy);
    }

    /**
     * @return Number of evicted frames.
     */
    inline std::size_t evicted_frames() const
    {
        return _evicted_frames;
    }

  private:
    StorageManager &_space_manager;
    std::unique_ptr<BufferReplacementStrategy> _replacement_strategy;

    std::vector<Frame> _frames;
    std::vector<FrameInformation> _frame_information;
    std::size_t _pin_sequence = 0u;
    std::size_t _evicted_frames = 0u;

    /**
     * Writes all dirty pages from memory to disk.
     */
    void flush();

    /**
     * Lookup for frame information for a specific page.
     * The frame information stores information like pin
     * history, pinned page for a frame.
     *
     * @param page_id Id of the page.
     * @return An iterator to the frame information or end() if the frame was not found.
     */
    std::vector<FrameInformation>::iterator frame_information(const Page::page_id page_id);

    /**
     * Find a frame which should be occupied for a new page request.
     * The target frame can be a) a free frame or b) a frame holding
     * another, not pinned page.
     *
     * @return Index for the frame
     */
    std::size_t find_frame();
};
} // namespace beedb::disk