/*------------------------------------------------------------------------------*
 * Architecture & Implementation of DBMS                                        *
 *------------------------------------------------------------------------------*
 * Copyright 2020 Databases and Information Systems Group TU Dortmund           *
 * Visit us at                                                                  *
 *             http://dbis.cs.tu-dortmund.de/cms/en/home/                       *
 *                                                                              *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS      *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,  *
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL      *
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR         *
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,        *
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR        *
 * OTHER DEALINGS IN THE SOFTWARE.                                              *
 *                                                                              *
 * Authors (alphabetical list):                                                 *
 *          Maximilian Berens   <maximilian.berens@tu-dortmund.de>              *
 *          Roland Kühn         <roland.kuehn@cs.tu-dortmund.de>                *
 *          Jan Mühlig          <jan.muehlig@tu-dortmund.de>                    *
 *------------------------------------------------------------------------------*
 */

#pragma once

#include "operator_interface.h"
#include "tuple_buffer.h"
#include <disk/buffer_manager.h>
#include <disk/page.h>
#include <index/index_interface.h>
#include <memory>
#include <queue>
#include <set>
#include <table/table.h>
#include <table/table_disk_manager.h>
#include <table/tuple.h>
#include <vector>

namespace beedb::execution
{
/**
 * Key (range) that have to be looked up in the.
 * May be a range or a single key.
 */
class KeyRange
{
  public:
    KeyRange(const std::int64_t single_key) : _from(single_key), _to(std::numeric_limits<std::int64_t>::max())
    {
    }

    KeyRange(const std::int64_t from, const std::int64_t to) : _from(from), _to(to)
    {
    }

    ~KeyRange() = default;

    bool is_single_key() const
    {
        return _to == std::numeric_limits<std::int64_t>::max();
    }
    std::int64_t single_key() const
    {
        return _from;
    }
    std::int64_t from() const
    {
        return _from;
    }
    std::int64_t to() const
    {
        return _to;
    }

    bool operator<(const KeyRange &other) const
    {
        return _from < other._from;
    }

  private:
    const std::int64_t _from;
    const std::int64_t _to;
};

/**
 * Takes an index and keys to be looked up in the index
 * and scans only over pages found in the index instead
 * of scanning all pages from the table.
 */
class IndexScanOperator final : public OperatorInterface
{
  public:
    IndexScanOperator(const std::uint32_t scan_page_limit, const table::Schema &schema,
                      disk::BufferManager &buffer_manager, table::TableDiskManager &table_disk_manager,
                      std::set<KeyRange> &key_ranges, std::shared_ptr<index::IndexInterface> index);
    virtual ~IndexScanOperator() = default;

    virtual void open();
    virtual util::optional<table::Tuple> next();
    virtual void close();

    virtual const table::Schema &schema() const
    {
        return _schema;
    }

  private:
    const std::uint32_t _scan_page_limit;
    const table::Schema _schema;
    disk::BufferManager &_buffer_manager;
    table::TableDiskManager &_table_disk_manager;
    std::set<KeyRange> _key_ranges;
    std::shared_ptr<index::IndexInterface> _index;

    std::queue<disk::Page::page_id> _pages_to_scan;
    std::vector<disk::Page::page_id> _pinned_pages;

    TupleBuffer _buffer;
};
} // namespace beedb::execution