/*------------------------------------------------------------------------------*
 * Architecture & Implementation of DBMS                                        *
 *------------------------------------------------------------------------------*
 * Copyright 2020 Databases and Information Systems Group TU Dortmund           *
 * Visit us at                                                                  *
 *             http://dbis.cs.tu-dortmund.de/cms/en/home/                       *
 *                                                                              *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS      *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,  *
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL      *
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR         *
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,        *
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR        *
 * OTHER DEALINGS IN THE SOFTWARE.                                              *
 *                                                                              *
 * Authors (alphabetical list):                                                 *
 *          Maximilian Berens   <maximilian.berens@tu-dortmund.de>              *
 *          Roland Kühn         <roland.kuehn@cs.tu-dortmund.de>                *
 *          Jan Mühlig          <jan.muehlig@tu-dortmund.de>                    *
 *------------------------------------------------------------------------------*
 */

#pragma once
#include <memory>

#include <table/schema.h>
#include <table/tuple.h>

namespace beedb::execution
{

/**
 * Interface for predicate matcher. The predicate matcher
 * takes one or two tuples and compares them with a constant
 * or a value from another tuple.
 */
class PredicateMatcherInterface
{
  public:
    enum Comparison
    {
        EQ,
        LE,
        LT,
        GE,
        GT,
        NEQ
    };
    virtual ~PredicateMatcherInterface() = default;
    virtual bool matches(const table::Tuple &tuple) = 0;
    virtual bool matches(const table::Tuple &left, const table::Tuple &right) = 0;
};

/**
 * The given tuple will always match.
 */
class AlwaysTrueMatcher final : public PredicateMatcherInterface
{
  public:
    AlwaysTrueMatcher() = default;

    virtual ~AlwaysTrueMatcher() = default;

    virtual bool matches(const table::Tuple &)
    {
        return true;
    }

    virtual bool matches(const table::Tuple &, const table::Tuple &)
    {
        return true;
    }
};

/**
 * Takes two predicate matchers p1 and p2 and connects them using AND.
 * Tuple matches when p1 AND p2 matches.
 */
class AndMatcher final : public PredicateMatcherInterface
{
  public:
    AndMatcher(std::unique_ptr<PredicateMatcherInterface> left, std::unique_ptr<PredicateMatcherInterface> right)
        : _left(std::move(left)), _right(std::move(right))
    {
    }

    virtual ~AndMatcher() = default;

    virtual bool matches(const table::Tuple &tuple)
    {
        return _left->matches(tuple) && _right->matches(tuple);
    }

    virtual bool matches(const table::Tuple &left, const table::Tuple &right)
    {
        return _left->matches(left, right) && _right->matches(left, right);
    }

  private:
    std::unique_ptr<PredicateMatcherInterface> _left;
    std::unique_ptr<PredicateMatcherInterface> _right;
};

/**
 * Takes two predicate matchers p1 and p2 and connects them using OR.
 * Tuple matches when p1 OR p2 matches.
 */
class OrMatcher final : public PredicateMatcherInterface
{
  public:
    OrMatcher(std::unique_ptr<PredicateMatcherInterface> left, std::unique_ptr<PredicateMatcherInterface> right)
        : _left(std::move(left)), _right(std::move(right))
    {
    }

    virtual ~OrMatcher() = default;

    virtual bool matches(const table::Tuple &tuple)
    {
        return _left->matches(tuple) || _right->matches(tuple);
    }

    virtual bool matches(const table::Tuple &left, const table::Tuple &right)
    {
        return _left->matches(left, right) || _right->matches(left, right);
    }

  private:
    std::unique_ptr<PredicateMatcherInterface> _left;
    std::unique_ptr<PredicateMatcherInterface> _right;
};

/**
 * Compares a specific column in a tuple with a constant.
 */
template <PredicateMatcherInterface::Comparison C> class AttributeValueMatcher final : public PredicateMatcherInterface
{
  public:
    AttributeValueMatcher(const table::Schema::ColumnIndexType schema_index, const table::Value value)
        : _schema_index(schema_index), _value(value)
    {
    }

    ~AttributeValueMatcher() = default;

    virtual bool matches(const table::Tuple &tuple)
    {
        if constexpr (C == EQ)
        {
            return tuple.get(_schema_index) == _value;
        }
        else if constexpr (C == LE)
        {
            return tuple.get(_schema_index) <= _value;
        }
        else if constexpr (C == LT)
        {
            return tuple.get(_schema_index) < _value;
        }
        else if constexpr (C == GE)
        {
            return tuple.get(_schema_index) >= _value;
        }
        else if constexpr (C == GT)
        {
            return tuple.get(_schema_index) > _value;
        }
        else
        {
            return tuple.get(_schema_index) != _value;
        }
    }

    virtual bool matches(const table::Tuple &, const table::Tuple &)
    {
        return false;
    }

  protected:
    const table::Schema::ColumnIndexType _schema_index;
    const table::Value _value;
};

/**
 * Compares two columns in a tuple or one column of two tuples.
 */
template <PredicateMatcherInterface::Comparison C> class AttributeMatcher final : public PredicateMatcherInterface
{
  public:
    AttributeMatcher(const table::Schema::ColumnIndexType schema_index_left,
                     const table::Schema::ColumnIndexType schema_index_right)
        : _schema_index_left(schema_index_left), _schema_index_right(schema_index_right)
    {
    }

    ~AttributeMatcher() = default;

    virtual bool matches(const table::Tuple &tuple)
    {
        if constexpr (C == EQ)
        {
            return tuple.get(_schema_index_left) == tuple.get(_schema_index_right);
        }
        else if constexpr (C == LE)
        {
            return tuple.get(_schema_index_left) <= tuple.get(_schema_index_right);
        }
        else if constexpr (C == LT)
        {
            return tuple.get(_schema_index_left) < tuple.get(_schema_index_right);
        }
        else if constexpr (C == GE)
        {
            return tuple.get(_schema_index_left) >= tuple.get(_schema_index_right);
        }
        else if constexpr (C == GT)
        {
            return tuple.get(_schema_index_left) > tuple.get(_schema_index_right);
        }
        else
        {
            return tuple.get(_schema_index_left) != tuple.get(_schema_index_right);
        }
    }

    virtual bool matches(const table::Tuple &left, const table::Tuple &right)
    {
        if constexpr (C == EQ)
        {
            return left.get(_schema_index_left) == right.get(_schema_index_right);
        }
        else if constexpr (C == LE)
        {
            return left.get(_schema_index_left) <= right.get(_schema_index_right);
        }
        else if constexpr (C == LT)
        {
            return left.get(_schema_index_left) < right.get(_schema_index_right);
        }
        else if constexpr (C == GE)
        {
            return left.get(_schema_index_left) >= right.get(_schema_index_right);
        }
        else if constexpr (C == GT)
        {
            return left.get(_schema_index_left) > right.get(_schema_index_right);
        }
        else
        {
            return left.get(_schema_index_left) != right.get(_schema_index_right);
        }
    }

    table::Schema::ColumnIndexType left_index() const
    {
        return _schema_index_left;
    }

    table::Schema::ColumnIndexType right_index() const
    {
        return _schema_index_right;
    }

  protected:
    const table::Schema::ColumnIndexType _schema_index_left;
    const table::Schema::ColumnIndexType _schema_index_right;
};
} // namespace beedb::execution