/*------------------------------------------------------------------------------*
 * Architecture & Implementation of DBMS                                        *
 *------------------------------------------------------------------------------*
 * Copyright 2020 Databases and Information Systems Group TU Dortmund           *
 * Visit us at                                                                  *
 *             http://dbis.cs.tu-dortmund.de/cms/en/home/                       *
 *                                                                              *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS      *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,  *
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL      *
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR         *
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,        *
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR        *
 * OTHER DEALINGS IN THE SOFTWARE.                                              *
 *                                                                              *
 * Authors (alphabetical list):                                                 *
 *          Maximilian Berens   <maximilian.berens@tu-dortmund.de>              *
 *          Roland Kühn         <roland.kuehn@cs.tu-dortmund.de>                *
 *          Jan Mühlig          <jan.muehlig@tu-dortmund.de>                    *
 *------------------------------------------------------------------------------*
 */

#pragma once

#include "expression/attribute.h"
#include <cassert>
#include <exception/logical_exception.h>
#include <memory>
#include <string>
#include <variant>
#include <vector>

namespace beedb::expression
{

/**
 * This represents the result of a predicate. We need ternary logic, since all
 * predicates, that involve attributes, can not be resolved statically.
 * Therefore, only constants can evaluate to either true or false, everything
 * else will result in std::nullopt, i.e. no result.
 *
 * This can be used to statically evaluate literal predicates at plan-creation
 * time.
 *
 * TODO: finish implementation, when this feature is needed.
 */
using TernaryValue = std::optional<bool>;

/**
 * Operand of a predicate, type is defined by the parser.
 *
 * Node: Unequal types always compare to False. I.e. "1.0f == 1l" give False
 */
using Operand = std::variant<Attribute, std::string, std::int64_t, float, bool>;

struct Atom
{
    const Operand left;
    const Operand right;

    Atom(const Operand left_, const Operand right_) : left(left_), right(right_)
    {
    }
    virtual ~Atom() = default;

    virtual std::unique_ptr<Atom> clone(const Operand &left, const Operand &right) const = 0;

    std::unique_ptr<Atom> clone()
    {
        return this->clone(left, right);
    }

    virtual operator TernaryValue() const = 0;

    operator std::string() const
    {
        std::string left_string;
        std::visit(OperandStringifier{left_string}, left);
        std::string right_string;
        std::visit(OperandStringifier{right_string}, right);
        return left_string + " " + operatorString() + " " + right_string;
    }

  protected:
    virtual std::string operatorString() const = 0;

    bool semantic_check() const
    {
        if (std::holds_alternative<Attribute>(left) || std::holds_alternative<Attribute>(right))
        {
            return false;
        }
        else if (left.index() != right.index())
        {
            throw exception::LogicalException("No support for operands of different literal types!");
        }
        return true;
    }

    struct OperandStringifier
    {
        // helper struct for visitor pattern.
        // implements string conversion of atoms by calling the
        // proper conversion operation on the operands
        std::string &rep;

        // depending on the variant type,
        // c++ picks the appropiate call operator

        // NOTE: 'overload' might be of use here
        // https://www.bfilipek.com/2018/06/variant.html#overload
        void operator()(std::int64_t const &val)
        {
            rep = std::to_string(val); // converts operand to std::string
        }
        void operator()(float const &val)
        {
            rep = std::to_string(val); // converts operand to std::string
        }
        void operator()(bool const &val)
        {
            rep = std::to_string(std::int32_t(val)); // converts operand to std::string
        }
        void operator()(std::string const &val)
        {
            rep = "\'" + val + "\'"; // converts operand to std::string
        }
        void operator()(Attribute const &val)
        {
            rep = val; // casts to string
        }
    };
};

struct EQ : public Atom
{
    EQ(const Operand left_, const Operand right_) : Atom{left_, right_}
    {
    }

    std::unique_ptr<Atom> clone(const Operand &left_, const Operand &right_) const override
    {
        return std::make_unique<EQ>(left_, right_);
    }

    operator TernaryValue() const override
    {
        return semantic_check() ? std::optional(left == right) : std::nullopt;
    }
    std::string operatorString() const override
    {
        return "=";
    }
};

struct NEQ : public Atom
{
    NEQ(const Operand left_, const Operand right_) : Atom{left_, right_}
    {
    }

    std::unique_ptr<Atom> clone(const Operand &left_, const Operand &right_) const override
    {
        return std::make_unique<NEQ>(left_, right_);
    }

    operator TernaryValue() const override
    {
        return semantic_check() ? std::optional(left != right) : std::nullopt;
    }
    std::string operatorString() const override
    {
        return "!=";
    }
};

struct LT : public Atom
{
    LT(const Operand left_, const Operand right_) : Atom{left_, right_}
    {
    }

    std::unique_ptr<Atom> clone(const Operand &left_, const Operand &right_) const override
    {
        return std::make_unique<LT>(left_, right_);
    }

    operator TernaryValue() const override
    {
        return semantic_check() ? std::optional(left < right) : std::nullopt;
    }
    std::string operatorString() const override
    {
        return "<";
    }
};

struct GT : public Atom
{
    GT(const Operand left_, const Operand right_) : Atom{left_, right_}
    {
    }

    std::unique_ptr<Atom> clone(const Operand &left_, const Operand &right_) const override
    {
        return std::make_unique<GT>(left_, right_);
    }

    operator TernaryValue() const override
    {
        return semantic_check() ? std::optional(left > right) : std::nullopt;
    }
    std::string operatorString() const override
    {
        return ">";
    }
};

struct LE : public Atom
{
    LE(const Operand left_, const Operand right_) : Atom{left_, right_}
    {
    }

    std::unique_ptr<Atom> clone(const Operand &left_, const Operand &right_) const override
    {
        return std::make_unique<LE>(left_, right_);
    }

    operator TernaryValue() const override
    {
        return semantic_check() ? std::optional(left <= right) : std::nullopt;
    }
    std::string operatorString() const override
    {
        return "<=";
    }
};

struct GE : public Atom
{
    GE(const Operand left_, const Operand right_) : Atom{left_, right_}
    {
    }

    std::unique_ptr<Atom> clone(const Operand &left_, const Operand &right_) const override
    {
        return std::make_unique<GE>(left_, right_);
    }

    operator TernaryValue() const override
    {
        return semantic_check() ? std::optional(left >= right) : std::nullopt;
    }

    std::string operatorString() const override
    {
        return ">=";
    }
};

/**
 * Recursive definition of a predicate. Can either be an atom or a two-term
 * combination of other predicates:
 */
struct LogicalConnective;

using Predicate = std::variant<std::unique_ptr<LogicalConnective>, std::unique_ptr<Atom>, bool>;

/**
 * @brief The Expression struct defines the logical combination of two
 * Predicate objects.
 */
struct LogicalConnective
{
  protected:
    LogicalConnective(Predicate &&left_, Predicate &&right_) : left(std::move(left_)), right(std::move(right_))
    {
    }

  public:
    virtual ~LogicalConnective() = default;

    virtual std::unique_ptr<LogicalConnective> clone() const = 0;

    Predicate left;
    Predicate right;

    /**
     * @brief operator bool makes evaluation of expressions possible.
     */
    operator TernaryValue() const
    {
        TernaryValue left_result;
        std::visit(TermEvaluator{left_result}, left);
        TernaryValue right_result;
        std::visit(TermEvaluator{right_result}, right);

        return apply(left_result, right_result);
    }

    operator std::string() const
    {
        std::string left_result;
        std::visit(TermStringifier{left_result}, left);
        std::string right_result;
        std::visit(TermStringifier{right_result}, right);
        return "(" + left_result + " " + connector() + " " + right_result + ")";
    }

    struct TermEvaluator
    {
        // helper struct for visitor pattern
        // implements boolean evaluation of expressions by calling the
        // proper conversion operation (to bool-conversion)
        TernaryValue &result;

        // depending on the variant type,
        // c++ picks the appropiate call operator:
        void operator()(const std::unique_ptr<Atom> &atom)
        {
            result = static_cast<TernaryValue>(*atom); // converts atom to bool
        }
        void operator()(const std::unique_ptr<LogicalConnective> &expression)
        {
            result = static_cast<TernaryValue>(*expression);
        }

        void operator()(const bool truth_value)
        {
            result = truth_value;
        }
    };
    struct TermStringifier
    {
        // helper struct for visitor pattern
        // implements boolean evaluation of expressions by calling the
        // proper conversion operation (to bool-conversion)
        std::string &rep;

        // depending on the variant type,
        // c++ picks the appropiate call operator:
        void operator()(const std::unique_ptr<Atom> &atom)
        {
            rep = static_cast<std::string>(*atom); // converts atom to bool
        }
        void operator()(const std::unique_ptr<LogicalConnective> &expression)
        {
            rep = static_cast<std::string>(*expression);
        }

        void operator()(const bool truth_value)
        {
            rep = truth_value ? "True" : "False";
        }
    };

  protected:
    virtual TernaryValue apply(TernaryValue, TernaryValue) const = 0;
    virtual std::string connector() const = 0;
};

static inline std::string to_string(const Predicate &predicate)
{
    std::string str;
    std::visit(LogicalConnective::TermStringifier{str}, predicate);
    return str;
}

static inline TernaryValue statically_evaluate(const Predicate &predicate)
{
    TernaryValue result;
    std::visit(LogicalConnective::TermEvaluator{result}, predicate);
    return result;
}

static inline Predicate clone_predicate(const Predicate &predicate)
{
    if (std::holds_alternative<std::unique_ptr<LogicalConnective>>(predicate))
    {
        return std::get<std::unique_ptr<LogicalConnective>>(predicate)->clone();
    }
    else if (std::holds_alternative<std::unique_ptr<Atom>>(predicate))
    {
        return std::get<std::unique_ptr<Atom>>(predicate)->clone();
    }
    else if (std::holds_alternative<bool>(predicate))
    {
        return std::get<bool>(predicate);
    }

    assert(false && "Could not clone predicate");
    return {};
}

struct AND : public LogicalConnective
{
    AND(Predicate &&left_, Predicate &&right_) : LogicalConnective(std::move(left_), std::move(right_))
    {
    }

    std::unique_ptr<LogicalConnective> clone() const override
    {
        auto left_clone = clone_predicate(left);
        auto right_clone = clone_predicate(right);
        return std::make_unique<AND>(std::move(left_clone), std::move(right_clone));
    }

    TernaryValue apply(TernaryValue left_, TernaryValue right_) const override
    {
        if (left_.has_value())
        {
            if (!left_.value())
            { // if the lhs expression is False...
                // ... the whole statement is always False!
                return false;
            }
            return right_; // .. if true, AND is a neutral/redundant operation
        }
        else if (right_.has_value())
        {
            if (!right_.value())
            { // if the rhs expression is False...
                // ... the whole statement is always False!
                return false;
            }
            return left_; // .. if true, AND is a neutral/redundant operation
        }
        else
        {
            // when both values are std::nullopt, then....
            return std::nullopt; // ..this statement is not statically evaluable!
        }
    }
    std::string connector() const override
    {
        return "&&";
    }
};

struct OR : public LogicalConnective
{
    OR(Predicate &&left_, Predicate &&right_) : LogicalConnective(std::move(left_), std::move(right_))
    {
    }

    std::unique_ptr<LogicalConnective> clone() const override
    {
        auto left_clone = clone_predicate(left);
        auto right_clone = clone_predicate(right);
        return std::make_unique<OR>(std::move(left_clone), std::move(right_clone));
    }

    TernaryValue apply(TernaryValue left_, TernaryValue right_) const override
    {
        if (left_.has_value())
        {
            if (left_.value())
            { // if the lhs expression is True...
                // ... the whole statement is always True!
                return true;
            }
            return right_; // .. if False, OR is a neutral/redundant operation
        }
        else if (right_.has_value())
        {
            if (right_.value())
            { // if the rhs expression is True...
                // ... the whole statement is always true!
                return true;
            }
            return left_; // .. if False, OR is a neutral/redundant operation
        }
        else
        {
            // when both values are std::nullopt, then....
            return std::nullopt; // ..this statement is not statically evaluable!
        }
    }
    std::string connector() const override
    {
        return "||";
    }
};

class PredicateAnalyzer
{
  public:
    static expression::Attributes attributes(const Predicate &predicate);
    static bool contains_range_predicate(const Predicate &predicate);
    static bool contains_not_equals_predicate(const Predicate &predicate);
};
} // namespace beedb::expression