/*------------------------------------------------------------------------------*
 * Architecture & Implementation of DBMS                                        *
 *------------------------------------------------------------------------------*
 * Copyright 2020 Databases and Information Systems Group TU Dortmund           *
 * Visit us at                                                                  *
 *             http://dbis.cs.tu-dortmund.de/cms/en/home/                       *
 *                                                                              *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS      *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,  *
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL      *
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR         *
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,        *
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR        *
 * OTHER DEALINGS IN THE SOFTWARE.                                              *
 *                                                                              *
 * Authors (alphabetical list):                                                 *
 *          Maximilian Berens   <maximilian.berens@tu-dortmund.de>              *
 *          Roland Kühn         <roland.kuehn@cs.tu-dortmund.de>                *
 *          Jan Mühlig          <jan.muehlig@tu-dortmund.de>                    *
 *------------------------------------------------------------------------------*
 */

#pragma once

#include "parser/query_parser.h"
#include <memory>

// hsql:
#include <SQLParser.h>

namespace beedb::parser
{
class HSQLParser : public QueryParser
{
  public:
    HSQLParser(const std::string &query);
    virtual ~HSQLParser() = default;

  protected:
  private:
    // implementation of parser functionality (see QueryParser class):
    const SelectExpression extractSELECT_impl() const override;
    const std::optional<FromExpression> extractFROM_impl() const override;
    const std::optional<WhereExpression> extractWHERE_impl() const override;
    const std::optional<GroupByExpression> extractGROUPBY_impl() const override;
    const std::optional<OrderByExpression> extractORDERBY_impl() const override;
    const std::optional<LimitExpression> extractLIMIT_impl() const override;

    virtual const std::optional<CreateTableExpression> extractCREATE_TABLE_impl() const override;
    virtual const std::optional<CreateIndexExpression> extractCREATE_INDEX_impl() const override;

    virtual const std::optional<InsertExpression> extractINSERT_impl() const override;

    virtual const UpdateExpression extractUPDATE_impl() const override;

    /// HELPER FUNCTIONS:
    /**
     * @brief extract_From_impl evaluates a TableRef expression.
     * Due to how joins are represented, this has to be done recursively.
     *
     * This function has side-effects (on fexp param).
     * @param fexp holds the result (and more)
     * @param from_expr the expression to evaluate
     */
    std::unique_ptr<FromDescr> extractFROM_impl(const hsql::TableRef *table_ref) const;

    /**
     * @brief extract_predicate converts a predicate hsql::Expr* to a predicate in the database internal representation.
     * Is called recursively for logical connected sub-predicates. Makes use of the build_operand helper, defined below.
     * @param expression
     * @return a std::variant<> of pointers
     */
    expression::Predicate extract_predicate(const hsql::Expr *expression) const;
    /**
     * @brief build_operand extracts operands of an atom
     * @param expression
     * @return a std::variant<> of values
     */
    const expression::Operand build_operand(const hsql::Expr *expression) const;

    /**
     * @brief build_attribute is a helper that builds a expression::Attribute.
     * This is done by checking for existence of alias, table references etc.
     * @param h_expr
     * @return
     */
    expression::Attribute build_attribute(const hsql::Expr *h_expr) const;
    expression::AttributeOrigin determine_function_type(const std::string &function_name) const;

    // hsql specific members:
    hsql::SQLParserResult _hsql_result;
    const hsql::SQLStatement *_hsql_statement;
};
} // namespace beedb::parser