/*------------------------------------------------------------------------------*
 * Architecture & Implementation of DBMS                                        *
 *------------------------------------------------------------------------------*
 * Copyright 2020 Databases and Information Systems Group TU Dortmund           *
 * Visit us at                                                                  *
 *             http://dbis.cs.tu-dortmund.de/cms/en/home/                       *
 *                                                                              *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS      *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,  *
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL      *
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR         *
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,        *
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR        *
 * OTHER DEALINGS IN THE SOFTWARE.                                              *
 *                                                                              *
 * Authors (alphabetical list):                                                 *
 *          Maximilian Berens   <maximilian.berens@tu-dortmund.de>              *
 *          Roland Kühn         <roland.kuehn@cs.tu-dortmund.de>                *
 *          Jan Mühlig          <jan.muehlig@tu-dortmund.de>                    *
 *------------------------------------------------------------------------------*
 */

#pragma once

#include "expression/operator.h"
#include "plan/graph/basic_graph.h"

#include <functional>

namespace beedb::plan::logical
{

/**
 * Type of node identifiers. Is supposed to be unique for every operator instance within a plan.
 * This type must be usable as a key in std::map<KEY_TYPE, ...>
 */
using NodeIDType = std::string;

/**
 * @brief The LogicalPlan class This is a (directed) graph of logical operators with schema- and data flow indicated as
 * edges.
 *
 * Nodes are (logical) Operators and Edges contain the schema produced by an operator. Nodes contain a dependency, that
 * has to be satisfied by incoming edges.
 *
 * It is created by consuming a query object and producing the canonical (and probably inefficient) query plan.
 *
 * In a canonical plan, all occuring relations are first combined via cross-product, or, if explicitly defined by the
 * user, as an inner join. Aggregations/Group by, selections and projections are defined "on top" of this cross-product.
 * The resulting (unoptimized) graph is supposed to be semantically identical to the query. This is enforced by the
 * notion of 'required schemas'. I.e. a selection, say, "A < 5", expects to find Attribute A in the Attributes of
 * incoming edges. Operators can not reduce (but extend or replace) the list of input-attributes in order to produce the
 * output schema, except for projection-operations.
 *
 * The plan is a graph and implements semantical consistency check's (see 'required schemas').
 * Semantic consistency is ensured, when all operator requirements are 'producable' by child operators, over the
 * attributes set on incoming edges.
 *
 * Operators can check for additional conditions, for example the Scan operators requires a given table to be present in
 * the database.
 *
 *
 * Note that Operator's and it's components are all read only, i.e. const. The plan itself is created operator by
 * operator.
 *
 * When performing plan optimizations, a plan is supposed to be treated as const and has to be recreated (based on an
 * unoptimized one).
 */
class Plan : public beedb::util::Graph<expression::Operator, expression::OutgoingSchema, NodeIDType>
{
  public:
    Plan() = default;

    /**
     * @brief producedSchema Generate schema that the result of this generator has.
     *
     * This is the union of attributes produced by operator with the id nid and all incoming attributes.
     * @param nid
     * @return the union of incoming attributes and additions of this operator
     */
    expression::OutgoingSchema produced_attributes(const NodeIDType &nid) const;

    bool requirements_satisfied(const NodeIDType node_id) const; // TODO: make use of phantom types

    // convenience function. This can be done directly via "plan[Edge]"
    // operator, but notation is then a little cluttered...
    void connect(const NodeIDType source_id, const NodeIDType sink_id)
    {
        (*this)[EdgeID{source_id, sink_id}] = this->produced_attributes(source_id);
    }

    /**
     * @brief descendAndApply basic "descend" function, that calls a function on a node and recursivly on all it's
     * children. Used to retrieve information from an existing plan, without altering it's state.
     *
     * @param starting_node where to start the descend
     * @param fkt this is called on the starting_node and all its children. Is expected to have state and side-effects
     * outside of logical::Plan.
     */
    void descend_and_apply(const NodeIDType &starting_node, std::function<void(const NodeIDType &)> fkt) const
    {
        fkt(starting_node);
        for (const auto &child : _incoming_node_ids.at(starting_node))
        {
            descend_and_apply(child, fkt);
        }
    }

    // typedefs for better readability:
    using NodeMapContent = NodeData::value_type; // aka. a std::pair<NodeIDType, std::unique_ptr<expression::Operator>>,
    using NodeRecreationFunction = std::function<NodeMapContent(const NodeIDType &)>;

    const NodeIDType descend_and_recreate(const Plan &other, const NodeIDType &others_node,
                                          NodeRecreationFunction recreate)
    {
        // recreate others_node and insert it into this plan. The id of the new node is stored in "node"
        const NodeIDType &node = this->insert(recreate(others_node));

        // traverse other_node's children:
        for (const auto &others_child : other._incoming_node_ids.at(others_node))
        {
            // recursively recreate all nodes of other:
            const auto &new_child = descend_and_recreate(other, others_child, recreate);
            // all of chilren's children should be created by now, making correct schema interence possible:
            this->connect(new_child, node);
        }

        return node;
    }

    /**
     * @brief insertBelow inserts a child below the specified parent. If there is already a child linked to the parent,
     * we insert in between the old child and the parent, preserving the flow of schema/data. Edge information is
     * recalculated
     * @param new_child
     * @param parent
     */
    void insert_below(const NodeIDType new_child, const NodeIDType parent)
    {
        // note: children send data to their parents, from leaves to the root

        if (_incoming_node_ids[parent].size() > 1)
        {
            std::cerr << "ERROR(logical::Plan): I am supposed to replace more then one child with a single new child ("
                      << new_child << "). This is undefined behaviour, no action is performed! (Parent node is "
                      << parent << ")" << std::endl;
            return;
        }
        else if (_incoming_node_ids[parent].size() == 1)
        {
            // in case there already exists a child below the parent,
            // we insert the new child between the parent and the old child!

            // first, we remove the existing edge from the parent's old child to the parent:
            auto &old_child = _incoming_node_ids[parent][0];
            (*this)[EdgeID{old_child, parent}].remove();

            // connect old child to new child (aka. replacing the removed edge):
            connect(old_child, new_child);
        }

        connect(new_child, parent); // in any valid case, we connect the new child to the parent
    }

    //        void generateDotFile(std::string& filepath); // TODO

    inline NodeIDType get_new_operator_ID(const std::string &name = "operator")
    {
        _last_operator_id = std::to_string(++_operator_count) + "_" + name;
        return _last_operator_id;
    }

    inline NodeIDType last_added_node() const
    {
        return _last_operator_id;
    }

    void print_table() const;

    /**
     * @brief findRoot returns the root node, if one can be determined
     * @return is std::nullopt, if no root is found
     */
    std::optional<NodeIDType> find_root() const
    {
        assert(_incoming_node_ids.size() == _node_data.size());
        assert(_node_data.size() == _outgoing_node_ids.size());

        if (_node_data.size() == 1)
        {
            return _node_data.begin()->first;
        }

        // the root is supposed to be the only node, that has only incoming nodes.
        std::optional<NodeIDType> root = std::nullopt;
        for (const auto &[node_id, _] : _node_data)
        {
            assert(_incoming_node_ids.find(node_id) != _incoming_node_ids.end());
            assert(_outgoing_node_ids.find(node_id) != _outgoing_node_ids.end());
            if (_incoming_node_ids.at(node_id).size() > 0 && _outgoing_node_ids.at(node_id).size() == 0)
            {
                if (root.has_value())
                {
                    // there is another node that satisfies this criterion, we can not determine the root uniquely!
                    return std::nullopt;
                }
                root = node_id;
            }
        }
        // root remains empty, if no node fulfilling the criterion is found
        return root;
    }

    /**
     * @brief is_proper_plan_graph Consistency check, that determines, if this (directed) plan-graph is fully connected
     * by a single path.
     *
     * It traverses the graph "backwards", by starting at the root and "marking" all visited nodes.
     * If all nodes are visited in backwards-direction from the node, the graph is "proper".
     *
     * Note that this has nothing to do with operators, just the graph topology is considered.
     *
     * @return true, if graph is "proper"
     */
    bool is_proper_plan_graph() const
    {
        /// We traverse the graph "backwards", by starting at the root and "marking" all visited nodes
        /// Begin traversal:
        /// For all incoming nodes, do:
        ///     * remove id from "unvisited_nodes"
        ///     * call this procedure on children
        /// Note:
        /// * If we try to remove an id twice, we detected a cycle
        /// * if not all id's are consumed from "unvisited nodes", we have
        ///     "islands" of nodes that are are not connected to the root
        /// In both cases, the graph is "improper"

        auto root = find_root(); // makes a pass over all registered nodes/operators
        if (!root.has_value())
        {
            return false; // a plan without a well defined root is not "proper"
        }

        // put all node id's of the plan into a set:
        std::set<NodeIDType> unvisited_nodes;
        std::transform(_node_data.begin(), _node_data.end(), std::inserter(unvisited_nodes, unvisited_nodes.end()),
                       std::bind(&NodeData::value_type::first, std::placeholders::_1));

        bool is_cyclic = false;

        // create helper, that removes visited nodes from "unvisited_nodes"
        auto visit_node = [&](const NodeIDType &node) {
            if (unvisited_nodes.find(node) == unvisited_nodes.end())
            {
                is_cyclic = true;
            }
            else
            {
                unvisited_nodes.erase(node);
            }
        };

        descend_and_apply(root.value(), visit_node);

        return !is_cyclic && unvisited_nodes.empty();
    }

    //    /**
    //     * @brief clone
    //     * @return note that the returned plan-copy is mutable/not const!
    //     */
    //    Plan clone() const {
    //        return {};
    //    }
  private:
    unsigned _operator_count = 0;

    NodeIDType _last_operator_id;
};
} // namespace beedb::plan::logical