/*------------------------------------------------------------------------------*
 * Architecture & Implementation of DBMS                                        *
 *------------------------------------------------------------------------------*
 * Copyright 2020 Databases and Information Systems Group TU Dortmund           *
 * Visit us at                                                                  *
 *             http://dbis.cs.tu-dortmund.de/cms/en/home/                       *
 *                                                                              *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS      *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,  *
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL      *
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR         *
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,        *
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR        *
 * OTHER DEALINGS IN THE SOFTWARE.                                              *
 *                                                                              *
 * Authors (alphabetical list):                                                 *
 *          Maximilian Berens   <maximilian.berens@tu-dortmund.de>              *
 *          Roland Kühn         <roland.kuehn@cs.tu-dortmund.de>                *
 *          Jan Mühlig          <jan.muehlig@tu-dortmund.de>                    *
 *------------------------------------------------------------------------------*
 */

#pragma once
#include "schema.h"
#include <cassert>
#include <disk/page.h>
#include <ostream>
#include <string>
#include <vector>

#include <expression/attribute.h> // used to extract the schema in the logical format

namespace beedb::table
{
/**
 * Represents a table in the database.
 */
class Table
{
    friend std::ostream &operator<<(std::ostream &stream, const Table &table);

  public:
    Table(const std::int32_t id, const disk::Page::page_id page_id, const Schema &schema)
        : _id(id), _page_id(page_id), _schema(schema)
    {
    }

    ~Table() = default;

    /**
     * @return Name of the table.
     */
    inline const std::string &name() const
    {
        return _schema.table_name();
    }

    /**
     * @return Id of the table.
     */
    inline std::int32_t id() const
    {
        return _id;
    }

    /**
     * @return Id of the first data page.
     */
    inline disk::Page::page_id page_id() const
    {
        return _page_id;
    }

    /**
     * @return Id of the last data page (pages are like a linked list).
     */
    inline disk::Page::page_id last_page_id() const
    {
        return _last_page_id;
    }

    /**
     * Updates the last page of a table.
     *
     * @param page_id Id of the last page.
     */
    inline void last_page_id(const disk::Page::page_id page_id)
    {
        _last_page_id = page_id;
    }

    /**
     * @return Schema of the table.
     */
    inline const Schema &schema() const
    {
        return _schema;
    }

    Column &operator[](const std::size_t index)
    {
        return _schema[index];
    }

    Column &operator[](const expression::Attribute &attribute)
    {
        const auto index = _schema.column_index(attribute);
        assert(index.has_value());
        return _schema[index.value()];
    }

    const Column &operator[](const std::size_t index) const
    {
        return _schema[index];
    }

    const Column &operator[](const expression::Attribute &attribute) const
    {
        const auto index = _schema.column_index(attribute);
        assert(index.has_value());
        return _schema[index.value()];
    }

    /**
     * @brief operator beedb::plan::logical::Attributes
     * Conversion operator, that extracts the attributes of a table.
     * The resulting object is the representation of a "schema" (set) in the
     * logical plan.
     */
    operator beedb::expression::Attributes() const
    {
        return _schema.attributes();
    }

    /**
     * @return True, if the table is virtual aka not persisted.
     */
    inline bool is_virtual() const
    {
        return _id == -1;
    }

  private:
    const std::int32_t _id;
    const disk::Page::page_id _page_id;
    disk::Page::page_id _last_page_id = disk::Page::INVALID_PAGE_ID; // Will not be persisted
    Schema _schema;
};
} // namespace beedb::table