/*------------------------------------------------------------------------------*
 * Architecture & Implementation of DBMS                                        *
 *------------------------------------------------------------------------------*
 * Copyright 2020 Databases and Information Systems Group TU Dortmund           *
 * Visit us at                                                                  *
 *             http://dbis.cs.tu-dortmund.de/cms/en/home/                       *
 *                                                                              *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS      *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,  *
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL      *
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR         *
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,        *
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR        *
 * OTHER DEALINGS IN THE SOFTWARE.                                              *
 *                                                                              *
 * Authors (alphabetical list):                                                 *
 *          Maximilian Berens   <maximilian.berens@tu-dortmund.de>              *
 *          Roland Kühn         <roland.kuehn@cs.tu-dortmund.de>                *
 *          Jan Mühlig          <jan.muehlig@tu-dortmund.de>                    *
 *------------------------------------------------------------------------------*
 */

#pragma once

#include "schema.h"
#include "table.h"
#include "tuple.h"
#include <cstdint>
#include <disk/buffer_manager.h>
#include <disk/page.h>
#include <utility>
#include <vector>

namespace beedb::table
{
/**
 * The TableDiskManager specifies the interface between tables and the disk.
 */
class TableDiskManager
{
  public:
    TableDiskManager(disk::BufferManager &buffer_manager);
    ~TableDiskManager() = default;

    /**
     * Reads the content of a page and interprets it as tuples
     * for the given schema.
     *
     * @param page Page with raw content.
     * @param schema Schema for the tuples.
     * @return List of tuples stored at the given page.
     */
    std::vector<Tuple> read_rows(disk::Page *page, const Schema &schema);

    /**
     * Writes the tuple as raw content to a free page associated with the table.
     * The written page will be unpinned after write.
     *
     * @param table Table.
     * @param tuple Tuple to be written.
     */
    void add_row(Table &table, Tuple &&tuple);

    /**
     * Writes the tuple as raw content to a free page associated with the table.
     *
     * @param table Table.
     * @param schema Schema the returned tuple should hold.
     * @param tuple Tuple to be written.
     * @return The tuple that was written to the page. The page is still pinned!
     */
    Tuple add_row(Table &table, const Schema &schema, Tuple &&tuple);

    /**
     * Takes a modified tuple and writes it to the disk.
     *
     * @param tuple The modified tuple that should be written to disk.
     */
    void update_row(const Tuple &tuple);

  private:
    disk::BufferManager &_buffer_manager;

    /**
     * Scans for a page with enough free space for a new tuple.
     *
     * @param table Target table.
     *
     * @return Id of the page with free space.
     */
    disk::Page::page_id find_page_for_row(Table &table);

    /**
     * Adds a tuple to a free page.
     *
     * @param table Target table.
     * @param tuple Tuple to be written.
     * @return Page the tuple was written to and offset in bytes.
     */
    std::pair<disk::Page *, std::size_t> add_row(Table &table, Tuple &tuple);
};
} // namespace beedb::table