/*------------------------------------------------------------------------------*
 * Architecture & Implementation of DBMS                                        *
 *------------------------------------------------------------------------------*
 * Copyright 2020 Databases and Information Systems Group TU Dortmund           *
 * Visit us at                                                                  *
 *             http://dbis.cs.tu-dortmund.de/cms/en/home/                       *
 *                                                                              *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS      *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,  *
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL      *
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR         *
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,        *
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR        *
 * OTHER DEALINGS IN THE SOFTWARE.                                              *
 *                                                                              *
 * Authors (alphabetical list):                                                 *
 *          Maximilian Berens   <maximilian.berens@tu-dortmund.de>              *
 *          Roland Kühn         <roland.kuehn@cs.tu-dortmund.de>                *
 *          Jan Mühlig          <jan.muehlig@tu-dortmund.de>                    *
 *------------------------------------------------------------------------------*
 */

#include <exception/command_exception.h>
#include <io/command/custom_commands.h>
#include <regex>

using namespace beedb::io::command;

std::optional<beedb::io::Query> ShowCommand::execute(const std::string &parameters)
{
    std::regex tables_regex("tables", std::regex::icase);
    std::regex columns_regex("columns", std::regex::icase);
    std::regex indices_regex("indices", std::regex::icase);
    std::smatch match;

    if (std::regex_match(parameters, match, tables_regex))
    {
        return Query{"select name as table_name from system_tables order by id asc;"};
    }
    else if (std::regex_match(parameters, match, columns_regex))
    {
        return Query{"select t.name as table_name, c.name as column_name from system_columns c join system_tables t on "
                     "c.table_id = t.id order by t.id asc, c.id asc;"};
    }
    else if (std::regex_match(parameters, match, indices_regex))
    {
        return Query{"select t.name as table_name, c.name as column_name, i.name as index_name, i.is_unique as "
                     "is_unique, i.type_id as type from system_indices i join system_columns c on i.column_id = c.id "
                     "join system_tables t on c.table_id = t.id order by t.id asc, c.id asc, i.id asc;"};
    }

    throw exception::CommandSyntaxException("Parameter is not recognized!", help_str());
}

std::optional<beedb::io::Query> ExplainCommand::execute(const std::string &input)
{
    std::regex explain_regex("([graph|plan]*)\\s(.*)", std::regex::icase);
    std::smatch explain_regex_match;

    if (std::regex_match(input, explain_regex_match, explain_regex))
    {
        if (explain_regex_match[1].str() == "plan")
        {
            return Query{explain_regex_match[2], Query::ExplainLevel::Plan};
        }
        else if (explain_regex_match[1].str() == "graph")
        {
            return Query{explain_regex_match[2], Query::ExplainLevel::Graph};
        }
        else
        {
            throw exception::CommandSyntaxException("Unknown argument \'" + explain_regex_match[1].str() + "\'",
                                                    help_str());
        }
    }
    else if (input.size() == 0)
    {
        throw exception::CommandSyntaxException("No query specified!", help_str());
    }
    return Query{input, Query::ExplainLevel::Plan};
}

std::optional<beedb::io::Query> SetCommand::execute(const std::string &input)
{
    std::regex set_regex("(\\w+)\\s(.+)", std::regex::icase);
    std::regex value_regex("[0-9]+", std::regex::icase);
    std::smatch set_regex_match;
    std::smatch value_regex_match;

    if (std::regex_match(input, set_regex_match, set_regex))
    {
        const auto attribute_name = set_regex_match[1].str();
        const auto value = set_regex_match[2].str();

        if (!std::regex_match(value, value_regex_match, value_regex))
        {
            throw exception::CommandException("Options can only be set to numerical values!");
        }

        // Setting attribute to value....
        std::optional<Config::ConfigValue> old_value;
        if (_config.contains(attribute_name))
        {
            old_value = _config[attribute_name];
        }

        _config.set(attribute_name, std::stoi(value));

        std::cout << "Setting option \'" << attribute_name << "\' to value " << value;
        if (old_value.has_value())
        {
            std::cout << " (was " << old_value.value() << ")";
        }
        std::cout << std::endl;
    }
    else if (input.size() > 0)
    {
        std::regex option_regex("(\\w+)", std::regex::icase);
        // TODO: set the default option instead of throwing an error!
        throw exception::CommandSyntaxException("No value specified for option \'" + input + "\'", help_str());
    }
    else
    {
        throw exception::CommandSyntaxException("No arguments specified!", help_str());
    }
    return std::nullopt;
}

std::optional<beedb::io::Query> GetCommand::execute(const std::string &input)
{
    std::regex get_regex("(\\w+)");
    std::smatch get_regex_match;

    if (std::regex_match(input, get_regex_match, get_regex))
    {
        // we try to find the option by name..
        if (_config.contains(input))
        {
            std::cout << input << ": " << _config[input] << std::endl;
        }
        else
        {
            throw exception::CommandException("Option not found!");
        }
    }
    else if (input.empty())
    {
        // when no argument was specified, we print all options:
        std::cout << static_cast<std::string>(_config) << std::endl;
    }
    else
    {
        throw exception::CommandException("\'" + input + "\' is not a valid option name!");
    }
    return std::nullopt;
}