/*------------------------------------------------------------------------------*
 * Architecture & Implementation of DBMS                                        *
 *------------------------------------------------------------------------------*
 * Copyright 2020 Databases and Information Systems Group TU Dortmund           *
 * Visit us at                                                                  *
 *             http://dbis.cs.tu-dortmund.de/cms/en/home/                       *
 *                                                                              *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS      *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,  *
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL      *
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR         *
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,        *
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR        *
 * OTHER DEALINGS IN THE SOFTWARE.                                              *
 *                                                                              *
 * Authors (alphabetical list):                                                 *
 *          Maximilian Berens   <maximilian.berens@tu-dortmund.de>              *
 *          Roland Kühn         <roland.kuehn@cs.tu-dortmund.de>                *
 *          Jan Mühlig          <jan.muehlig@tu-dortmund.de>                    *
 *------------------------------------------------------------------------------*
 */

#include <chrono>
#include <exception/exception.h>
#include <io/executor.h>
#include <iostream>
#include <parser/hsql_parser.h>
#include <parser/query_parser.h>
#include <plan/logical/builder.h>
#include <plan/optimizer/optimizer.h>
#include <plan/physical/builder.h>
#include <util/clock.h>

using namespace beedb::io;

ExecutionResult Executor::execute(const Query &query, std::function<void(const table::Schema &)> schema_callback,
                                  std::function<void(const table::Tuple &)> tuple_callback)
{
    std::chrono::milliseconds planning_time{}, execution_time{};
    try
    {
        ////////////////////////////////////////////////////////////////////////
        /// \brief parser this object offers all components of the posed query,
        /// to construct a (logical) plan.
        const parser::QueryParser &parser = parser::HSQLParser(query.query_string);

        util::Clock planning_clock{}; // we do not measure parsing time
        const auto before_query_evicted_frames = this->_database.buffer_manager().evicted_frames();

        ////////////////////////////////////////////////////////////////////////
        /// \brief logical_plan holds the canonical plan, created from the query.
        /// This plan has no optimizations at all.
        auto canonical_logical_plan = beedb::plan::logical::Builder::build(this->_database, parser);

        /// In case this is a SELECT query and optimization is enabled...
        if (parser.type() == parser::QueryParser::QueryType::SELECT &&
            !(this->_database.config()[Config::k_OptimizationDisableOptimization]))
        {
            ////////////////////////////////////////////////////////////////////////
            /// \brief optimizer holds a reference to the canonical plan and can create a new, optimized plan
            /// \brief optimized_plan might hold a "more efficient" plan, could be empty though
            plan::logical::Optimizer optimizer(this->_database, canonical_logical_plan);
            auto optimized_plan = optimizer.optimize();

            /// If there is no new, optimized plan, the unoptimized plan will be executed instead.
            const plan::logical::Plan &final_plan =
                !optimized_plan.is_empty() ? optimized_plan : canonical_logical_plan;

            if (this->_database.config()[Config::k_CheckFinalPlan])
            {
                assert(final_plan.is_proper_plan_graph()); // this check is not cheap.
            }
            ////////////////////////////////////////////////////////////////////////
            /// Explain the chosen plan, when requested:
            if (query.explain == Query::ExplainLevel::Graph)
            {
                if (!optimized_plan.is_empty())
                {
                    canonical_logical_plan.toConsole();
                }
                final_plan.toConsole();
            }
            else if (query.explain == Query::ExplainLevel::Plan)
            {
                final_plan.print_table();
            }

            ////////////////////////////////////////////////////////////////////////
            /// create the physical plan from the logical one:
            auto plan = beedb::plan::physical::Builder::build(this->_database, final_plan);
            planning_time = planning_clock.end();

            util::Clock execution_clock{};
            plan.execute(schema_callback, tuple_callback);
            execution_time = execution_clock.end();
        }
        else
        { // just execute the canonical plan otherwise
            ////////////////////////////////////////////////////////////////////////
            /// create the physical plan from the logical one:
            auto plan = beedb::plan::physical::Builder::build(this->_database, canonical_logical_plan);
            planning_time = planning_clock.end();

            util::Clock execution_clock{};
            plan.execute(schema_callback, tuple_callback);
            execution_time = execution_clock.end();
        }

        const auto final_evicted_frames =
            this->_database.buffer_manager().evicted_frames() - before_query_evicted_frames;
        return {planning_time, execution_time, final_evicted_frames};
    }
    catch (std::runtime_error &e)
    {
        std::cerr << "\033[0;31merror\033[0m> " << e.what() << std::endl;
        return {};
    }
    catch (beedb::exception::DatabaseException &e)
    {
        std::cerr << "\033[0;31merror\033[0m> " << e.what() << std::endl;
        return {};
    }
}

ExecutionResult Executor::execute(plan::physical::Plan &plan,
                                  std::function<void(const table::Schema &)> schema_callback,
                                  std::function<void(const table::Tuple &)> tuple_callback)
{
    std::chrono::milliseconds execution_time{};
    try
    {
        util::Clock execution_clock{};
        plan.execute(schema_callback, tuple_callback);
        execution_time = execution_clock.end();

        return {std::chrono::milliseconds{0}, execution_time, 0u};
    }
    catch (std::runtime_error &e)
    {
        std::cerr << "\033[0;31merror\033[0m> " << e.what() << std::endl;
        return {};
    }
    catch (beedb::exception::DatabaseException &e)
    {
        std::cerr << "\033[0;31merror\033[0m> " << e.what() << std::endl;
        return {};
    }
}