/*------------------------------------------------------------------------------*
 * Architecture & Implementation of DBMS                                        *
 *------------------------------------------------------------------------------*
 * Copyright 2020 Databases and Information Systems Group TU Dortmund           *
 * Visit us at                                                                  *
 *             http://dbis.cs.tu-dortmund.de/cms/en/home/                       *
 *                                                                              *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS      *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,  *
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL      *
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR         *
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,        *
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR        *
 * OTHER DEALINGS IN THE SOFTWARE.                                              *
 *                                                                              *
 * Authors (alphabetical list):                                                 *
 *          Maximilian Berens   <maximilian.berens@tu-dortmund.de>              *
 *          Roland Kühn         <roland.kuehn@cs.tu-dortmund.de>                *
 *          Jan Mühlig          <jan.muehlig@tu-dortmund.de>                    *
 *------------------------------------------------------------------------------*
 */

#include <fstream>
#include <io/file_executor.h>
#include <regex>

using namespace beedb::io;

void FileExecutor::import_file(const std::string &file_name)
{
    std::ifstream file(file_name);
    std::stringstream statement_stream;
    std::string line;
    std::vector<std::string> statements;

    std::regex comment_regex(".*#|.*--.*");
    std::regex long_comment_start_regex(".*\\/\\*");
    std::regex long_comment_end_regex(".*\\*\\/");
    std::regex statement_end_regex(".*;");
    std::smatch match;
    bool is_within_comment = false;

    while (std::getline(file, line))
    {
        if (std::regex_match(line, match, long_comment_end_regex))
        {
            is_within_comment = false;
            continue;
        }

        if (is_within_comment || std::regex_match(line, match, comment_regex))
        {
            continue;
        }

        if (std::regex_match(line, match, long_comment_start_regex))
        {
            is_within_comment = true;
            continue;
        }

        statement_stream << " " << line << std::flush;
        if (std::regex_match(line, match, statement_end_regex))
        {
            statements.push_back(statement_stream.str());
            statement_stream.str("");
            statement_stream.clear();
        }
    }

    const auto size = statements.size();
    auto current = std::size_t{0};
    for (const auto &statement : statements)
    {
        const auto c = ++current;
        std::cout << "\rExecuting " << c << " / " << size << " (" << std::uint16_t(100 / float(size) * float(c)) << "%)"
                  << std::flush;
        Executor::execute(Query{statement});
    }

    std::cout << "\r"
              << "\033[0;32m"
              << "Executed " << size << " statements\t\t"
              << " \033[0m" << std::endl;
}