/*------------------------------------------------------------------------------*
 * Architecture & Implementation of DBMS                                        *
 *------------------------------------------------------------------------------*
 * Copyright 2020 Databases and Information Systems Group TU Dortmund           *
 * Visit us at                                                                  *
 *             http://dbis.cs.tu-dortmund.de/cms/en/home/                       *
 *                                                                              *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS      *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,  *
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL      *
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR         *
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,        *
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR        *
 * OTHER DEALINGS IN THE SOFTWARE.                                              *
 *                                                                              *
 * Authors (alphabetical list):                                                 *
 *          Maximilian Berens   <maximilian.berens@tu-dortmund.de>              *
 *          Roland Kühn         <roland.kuehn@cs.tu-dortmund.de>                *
 *          Jan Mühlig          <jan.muehlig@tu-dortmund.de>                    *
 *------------------------------------------------------------------------------*
 */

#include <exception/command_exception.h>
#include <io/command/commander.h>
#include <io/command/custom_commands.h>
#include <io/user_console.h>
#include <iostream>
#include <regex>
#include <string>

using namespace beedb::io;

void UserConsole::wait_for_input()
{
    std::regex quit_regex("q|quit|:q|:quit|e|exit|:e|:exit", std::regex_constants::icase);
    std::cout << "Type 'q' or 'quit' to exit." << std::endl;

    command::Commander commander;
    commander.register_command("show", std::make_unique<command::ShowCommand>());
    commander.register_command("explain", std::make_unique<command::ExplainCommand>());
    commander.register_command("set", std::make_unique<command::SetCommand>(this->_database.config()));
    commander.register_command("get", std::make_unique<command::GetCommand>(this->_database.config()));

    while (true)
    {
        std::string user_input;

        std::cout << "beedb> " << std::flush;
        std::getline(std::cin, user_input);

        std::smatch match;
        if (std::regex_match(user_input, match, quit_regex))
        {
            std::cout << "Bye." << std::endl;
            return;
        }
        else if (commander.has_command_prefix(user_input))
        {
            try
            {
                std::optional<Query> query = commander.create_query(user_input);
                if (query.has_value())
                {
                    // Note: a command does not always require the execution of a query
                    PrintingExecutor::execute(query.value());
                }
            }
            catch (exception::DatabaseException &e)
            {
                std::cerr << "\033[0;31merror\033[0m> " << e.what() << std::endl;
            }
        }
        else
        {
            PrintingExecutor::execute(Query{user_input}); // false => no explain by default
        }
    }
}