/*------------------------------------------------------------------------------*
 * Architecture & Implementation of DBMS                                        *
 *------------------------------------------------------------------------------*
 * Copyright 2020 Databases and Information Systems Group TU Dortmund           *
 * Visit us at                                                                  *
 *             http://dbis.cs.tu-dortmund.de/cms/en/home/                       *
 *                                                                              *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS      *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,  *
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL      *
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR         *
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,        *
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR        *
 * OTHER DEALINGS IN THE SOFTWARE.                                              *
 *                                                                              *
 * Authors (alphabetical list):                                                 *
 *          Maximilian Berens   <maximilian.berens@tu-dortmund.de>              *
 *          Roland Kühn         <roland.kuehn@cs.tu-dortmund.de>                *
 *          Jan Mühlig          <jan.muehlig@tu-dortmund.de>                    *
 *------------------------------------------------------------------------------*
 */

#include <exception/logical_exception.h>
#include <plan/logical/builder.h>

using namespace beedb::plan::logical;

const Plan Builder::buildSelectPlan(Database &database, const parser::QueryParser &parser)
{

    /// FILL THE PLAN:
    /// We build the canonical, 'straight-forward' plan:
    ///
    /// 0. Projections ("root" of the plan/the last operator)
    /// 1. Selections/filters
    /// 2. Aggregations
    /// 3. Joins
    /// 4. Cross-products
    /// 5. TABLE nodes
    ///
    /// We also check semantics, but no optimization is done.

    logical::Plan plan;

    /// First: gather information from the parser:

    // get projection-components from parser object:
    const auto select_part = parser.extractSELECT();

    // get from-component from parser object
    const auto from_part = parser.extractFROM();

    // WHERE component
    const auto where_part = parser.extractWHERE();

    const auto groupby_part = parser.extractGROUPBY();

    const auto orderby_part = parser.extractORDERBY();

    // LIMIT component
    const auto limit_part = parser.extractLIMIT();

    AttributeToTableNameMapping attribute_mapping;
    TableAliasMapping table_references; // used as a dictionary: name->table_ptr
    std::vector<const table::Table *> source_tables;

    if (from_part.has_value())
    {
        /// Second: resolve alias-defintions and perform consistency checks:
        // build a mapping of alias and table descriptions. check for existing tables
        for (parser::TableDescr &source_desc : from_part.value().tables())
        {

            const auto &source = source_desc.first;
            const auto &alias = source_desc.second;

            if (!database.table_exists(source))
            {
                // EXCEPTION: The user has specified a table in the from part, that is not in the currently loaded
                // database
                throw exception::TableNotFoundException(source);
            }

            // check for proper disambiguation (via alias) in case of self-joins:
            const auto &table_name_iter = table_references.find(alias.value_or(source));

            if (table_name_iter != table_references.end())
            { // if name already exists...
                if (!alias)
                {
                    // not specifying an alias is an error, in this case
                    throw exception::MultipleTableReferences(source);
                }
                const auto is_alias_not_unique = table_references.find(alias.value()) != table_references.end();
                if (is_alias_not_unique)
                {
                    // alias has to be unique:
                    throw exception::MultipleTableReferences(source);
                }
            }

            // a table can be referenced with a user specified alias or its name
            table_references[alias.value_or(source)] = {database[source], static_cast<bool>(alias)};
            source_tables.push_back(database[source]); // for convenient, ordered table iteration
        }

        // check join predicates:
        expression::Attributes pred_attrs;
        for (const auto &predicate : from_part.value().predicates())
        {
            auto attributes = expression::PredicateAnalyzer::attributes(predicate);
            std::move(attributes.begin(), attributes.end(), std::back_inserter(pred_attrs));
        }

        if (where_part.has_value())
        {
            auto attributes = expression::PredicateAnalyzer::attributes(where_part.value());
            std::move(attributes.begin(), attributes.end(), std::back_inserter(pred_attrs));
        }

        for (const auto &pred_attr : pred_attrs)
        {
            populateAttributeToTableMappingAndCheck(attribute_mapping, table_references, pred_attr);
        }

        // TODO: semantic check for ORDER BY and GROUP BY attributes

        if (groupby_part.has_value() && groupby_part->size() > 1)
        {
            throw exception::MultipleGroupByException();
        }

        if (orderby_part.has_value())
        {
            for (const auto &ordered_attribute : orderby_part.value())
            {
                populateAttributeToTableMappingAndCheck(attribute_mapping, table_references, ordered_attribute);
            }
        }
    }

    // check projection attribute- and table references.
    // Note: final-result schema will be created in the last step.
    for (const auto &attr : select_part)
    {
        populateAttributeToTableMappingAndCheck(attribute_mapping, table_references, attr);
    }

    // Create table-, join- and crossproduct nodes:
    std::optional<expression::Predicate> predicate = std::nullopt;
    if (where_part.has_value())
    {
        predicate.emplace(resolveTableReferences(attribute_mapping, where_part.value()));
    }

    auto top_node = addFromNodes(database, plan, from_part.value().from_description->descr, attribute_mapping);

    if (where_part.has_value())
    {
        top_node = addSelection(plan, predicate.value(), top_node);
    }

    if (groupby_part.has_value())
    {
        top_node = addGroupBy(attribute_mapping, groupby_part.value(), plan, top_node);
    }

    if (orderby_part.has_value())
    {
        top_node = addOrderBy(attribute_mapping, orderby_part.value(), plan, top_node);
    }

    if (limit_part.has_value())
    {
        top_node = addLimit(limit_part.value(), plan, top_node);
    }

    addProjection(plan, select_part, attribute_mapping, top_node);

    // TODO: @Max Can this be removed? (seems to work without)
    // plan[plan.last_added_node()].outgoingNodes();

    return plan;
}

void Builder::populateAttributeToTableMappingAndCheck(AttributeToTableNameMapping &attribute_mapping,
                                                      const TableAliasMapping &table_references,
                                                      const expression::Attribute &attr)
{
    // TODO: implement functionality and remove this check:
    if (!(attr.origin == expression::AttributeOrigin::MIXED || attr.origin == expression::AttributeOrigin::PHYSICAL))
    {
        throw exception::LogicalException("No support for aggregation yet, sorry!");
    }

    if (attr.table.has_value())
    { // if an attribute is used with a table reference..
        // both table alias and actual table name should resolve:
        const auto &table_name_iter = table_references.find(attr.table.value());

        if (table_name_iter == table_references.end())
        {
            // EXCEPTION: the user used a table reference in the SELECT part, that can not be resolved to a table in
            // FROM part
            throw exception::TableNotFoundException(attr.table.value(), attr.combined_name);
        }

        // The table is available, find column name in that table:
        // Note: asterisk (i.e. "T.*") needs no check
        if (!attr.isAsterisk() && !(table_name_iter->second.first->schema().contains(attr.name)))
        {
            // i.e. "if attr doesnt exist..." :
            throw exception::CanNotResolveColumnException(attr.name, attr.combined_name);
        }
    }
    else if (!attr.isAsterisk())
    {
        // no table reference is given, we try to resolve it to tables mentioned in FROM part:
        // Asterisk "SELECT * FROM ..." needs no check..
        std::optional<std::string> last_table_name;
        for (auto const &[table_name_expression, table_ref_pair] : table_references)
        {
            if (table_ref_pair.first->schema().contains(attr.name))
            {
                const auto &new_table_name = table_name_expression;
                if (last_table_name.has_value() && new_table_name != last_table_name.value())
                {
                    throw exception::NoUniqueReferenceException(attr.combined_name, last_table_name.value(),
                                                                new_table_name);
                }
                last_table_name = new_table_name;
            }
        }

        if (!last_table_name.has_value())
        {
            throw exception::ColumnNotFoundException(attr.combined_name);
        }

        attribute_mapping[attr.combined_name] = last_table_name.value();
    }
    else if (table_references.size() == 0)
    {
        throw exception::CanNotResolveColumnException("*");
    }
}

beedb::expression::Predicate Builder::resolveTableReferences(const AttributeToTableNameMapping &attribute_mapping,
                                                             const expression::Predicate &pred)
{
    if (std::holds_alternative<std::unique_ptr<expression::LogicalConnective>>(pred))
    {
        const auto &logical_connective = std::get<std::unique_ptr<expression::LogicalConnective>>(pred);
        auto left = resolveTableReferences(attribute_mapping, logical_connective->left);
        auto right = resolveTableReferences(attribute_mapping, logical_connective->right);
        auto logical_connective_ptr = logical_connective.get();
        if (typeid(*logical_connective_ptr) == typeid(expression::AND))
        {
            return std::make_unique<expression::AND>(std::move(left), std::move(right));
        }
        else
        {
            return std::make_unique<expression::OR>(std::move(left), std::move(right));
        }
    }
    else if (std::holds_alternative<std::unique_ptr<expression::Atom>>(pred))
    {
        auto &atom = std::get<std::unique_ptr<expression::Atom>>(pred);

        // Here, we are going to move the atom, if it has an attribute, that holds no table reference.
        // The new atom is identical, except for the (new) table reference
        using NewAttribute = std::optional<expression::Operand>;
        const NewAttribute new_left = std::holds_alternative<expression::Attribute>(atom->left)
                                          ? (!std::get<expression::Attribute>(atom->left).table.has_value()
                                                 ? NewAttribute{resolveTableReferences(
                                                       attribute_mapping, std::get<expression::Attribute>(atom->left))}
                                                 : std::nullopt)
                                          : std::nullopt;
        const NewAttribute new_right =
            std::holds_alternative<expression::Attribute>(atom->right)
                ? (!std::get<expression::Attribute>(atom->right).table.has_value()
                       ? NewAttribute{resolveTableReferences(attribute_mapping,
                                                             std::get<expression::Attribute>(atom->right))}
                       : std::nullopt)
                : std::nullopt;

        if (new_left.has_value() || new_right.has_value())
        {
            auto left_operand = new_left.value_or(atom->left);
            auto right_operand = new_right.value_or(atom->right);

            return atom->clone(left_operand, right_operand);
        }
        else
        {
            return atom->clone();
        }
    }
    else if (std::holds_alternative<bool>(pred))
    {
        return std::get<bool>(pred);
    }

    assert(false && "NOT ALL CASES COVERED in logical/builder.cpp: Builder::resolveTableReferences()");
    return {};
}

const beedb::expression::Attribute Builder::resolveTableReferences(const AttributeToTableNameMapping &attribute_mapping,
                                                                   const beedb::expression::Attribute &attribute)
{
    return expression::Attribute::create(attribute, attribute_mapping.at(attribute.combined_name));
}

const NodeIDType Builder::addFromNodes(
    const Database &db, logical::Plan &plan,
    std::variant<parser::TableDescr, parser::JoinDescr, parser::CrossProductDescr> &from,
    const AttributeToTableNameMapping &attribute_mapping)
{
    logical::NodeIDType node_id = plan.get_new_operator_ID("FROM"); // for now, this works/is compatible
    switch (from.index())
    {
    case 0ul: { // TableDescr
        const auto t_d = std::get<parser::TableDescr>(from);
        const auto &schema_attr = db[t_d.first].schema().attributes();

        const auto attributes = expression::recreateAttributes(schema_attr, t_d.second.value_or(t_d.first));

        plan[node_id] = std::make_unique<expression::TableOperator>(attributes, t_d.first);
    }
    break;
    case 1ul: { // JoinDescr
        auto &join_description = std::get<parser::JoinDescr>(from);
        const auto &join_predicate = std::get<2>(join_description);

        plan[node_id] = std::make_unique<expression::JoinOperator>(
            resolveTableReferences(attribute_mapping, join_predicate),
            expression::JoinOperator::NestedLoopsJoin); // only NLJ's in the canonical plan

        // create and connect child nodes:
        const auto left_child = addFromNodes(db, plan, std::get<0>(join_description)->descr, attribute_mapping);

        plan.connect(left_child, node_id);

        const auto right_child = addFromNodes(db, plan, std::get<1>(join_description)->descr, attribute_mapping);
        plan.connect(right_child, node_id);
    }
    break;
    case 2ul: { // CrossProductDescr
        std::vector<logical::NodeIDType> child_ids;

        for (const auto &el : std::get<parser::CrossProductDescr>(from))
        {
            child_ids.push_back(addFromNodes(db, plan, el->descr, attribute_mapping));
        }
        // we build at least one CP. this will be the final top-node:
        plan[node_id] = std::make_unique<expression::CrossProductOperator>();

        plan.connect(child_ids[child_ids.size() - 2], node_id);
        plan.connect(child_ids[child_ids.size() - 1], node_id);

        // for each child beyond the second, we have to build another CP:
        for (auto i = static_cast<std::int32_t>(child_ids.size() - 3); i >= 0; i--)
        {
            // note: i-- will overflow into negatives on last iteration, breaking the loop condition

            const logical::NodeIDType parent_id = plan.get_new_operator_ID("FROM");
            plan[parent_id] = std::make_unique<expression::CrossProductOperator>();

            // connect a child to new CP:
            // note: child CP comes second, cross product tree "grows (upwards) to the left"
            plan.connect(child_ids[static_cast<std::size_t>(i)], parent_id);
            plan.connect(node_id, parent_id); // connect new CP to parent CP
            node_id = parent_id;              // set this, new CP as parent for next CP
        }
    }
    break;
    default:
        break;
    }

    return node_id;
}

const NodeIDType Builder::addSelection(logical::Plan &plan, const expression::Predicate &predicate,
                                       const logical::NodeIDType &parent_id)
{
    const auto selection_name = plan.get_new_operator_ID("SELECTION");
    plan[selection_name] = std::make_unique<expression::SelectionOperator>(clone_predicate(predicate));

    plan.connect(parent_id, selection_name);

    return selection_name;
}

const NodeIDType Builder::addGroupBy(const AttributeToTableNameMapping &attribute_mapping,
                                     const beedb::parser::GroupByExpression &groupbyexpr, Plan &plan,
                                     const NodeIDType &parent_id)
{
    // TODO: Add exception
    const auto groupby_name = plan.get_new_operator_ID("GROUPBY");
    expression::Attributes group_by_attributes;
    for (const auto &attribute : groupbyexpr)
    {
        if (attribute.table.has_value())
        {
            group_by_attributes.push_back(std::move(attribute));
        }
        else
        {
            group_by_attributes.push_back(resolveTableReferences(attribute_mapping, attribute));
        }
    }

    plan[groupby_name] = std::make_unique<expression::GroupByOperator>(group_by_attributes);
    plan.connect(parent_id, groupby_name);
    return groupby_name;
}

const NodeIDType Builder::addOrderBy(const AttributeToTableNameMapping &attribute_mapping,
                                     const beedb::parser::OrderByExpression &orderbyexpr, Plan &plan,
                                     const NodeIDType &parent_id)
{
    const auto orderby_name = plan.get_new_operator_ID("ORDERBY");
    expression::Attributes order_by_attributes;
    for (const auto &attribute : orderbyexpr)
    {
        if (attribute.table.has_value())
        {
            order_by_attributes.push_back(std::move(attribute));
        }
        else
        {
            order_by_attributes.push_back(resolveTableReferences(attribute_mapping, attribute));
        }
    }

    plan[orderby_name] = std::make_unique<expression::OrderByOperator>(order_by_attributes);
    plan.connect(parent_id, orderby_name);
    return orderby_name;
}

void Builder::addProjection(logical::Plan &plan, const std::vector<expression::Attribute> &attributes,
                            const AttributeToTableNameMapping &attribute_mapping, const logical::NodeIDType &parent_id)
{
    const auto projection_name = plan.get_new_operator_ID("PROJECTION");

    expression::Attributes projection;
    // create attributes of projection:
    for (const auto &attr : attributes)
    {
        if (attr.isAsterisk())
        {
            for (const auto &child_attr : plan.produced_attributes(parent_id))
            {
                // either resolve all attributes
                // or only retrieve attributes for a specified table:
                if ((!attr.table) || child_attr.table == attr.table)
                {
                    projection.push_back(child_attr);
                }
            }
        }
        else
        {
            if (!attr.table.has_value())
            {
                // resolve table reference
                projection.push_back(resolveTableReferences(attribute_mapping, attr));
            }
            else
            {
                projection.push_back(attr);
            }
        }
    }
    plan[projection_name] = std::make_unique<expression::ProjectionOperator>(projection);
    plan.connect(parent_id, projection_name);
}

const NodeIDType Builder::addLimit(const beedb::parser::LimitExpression &limit_expression,
                                   beedb::plan::logical::Plan &plan, const beedb::plan::logical::NodeIDType &parent_id)
{
    const auto limit_id = plan.get_new_operator_ID("LIMIT");
    plan[limit_id] = std::make_unique<expression::LimitOperator>(limit_expression.limit, limit_expression.offset);
    plan.connect(parent_id, limit_id);
    return limit_id;
}

const Plan Builder::buildCreateTablePlan(Database &database, const parser::QueryParser &parser)
{
    Plan plan;
    const auto create_table_name = plan.get_new_operator_ID("CREATE_TABLE");
    const auto create_table_expression = parser.extractCREATE_TABLE();
    if (create_table_expression.has_value() == false)
    {
        throw exception::CanNotCreateTableException();
    }

    const auto table_exists = database.table_exists(create_table_expression->table_name);
    if (table_exists && create_table_expression->if_not_exists == true)
    {
        return plan;
    }
    else if (table_exists)
    {
        throw exception::TableAlreadyExists(create_table_expression->table_name);
    }

    std::vector<std::string> column_names;
    std::vector<table::Type> column_types;
    std::vector<bool> column_nullables;

    for (const auto &column : create_table_expression->column_expressions)
    {
        column_names.push_back(column.column_name);
        column_types.push_back({column.type_id, column.length});
        column_nullables.push_back(column.is_null);
    }

    plan[create_table_name] =
        std::make_unique<expression::CreateTableOperator>(create_table_expression->table_name, std::move(column_names),
                                                          std::move(column_types), std::move(column_nullables));

    return plan;
}

const Plan Builder::buildCreateIndexPlan(Database &database, const parser::QueryParser &parser)
{
    Plan plan;

    const auto expression = parser.extractCREATE_INDEX();
    if (expression.has_value() == false)
    {
        throw exception::CanNotCreateIndexException();
    }

    if (database.table_exists(expression->table_name) == false)
    {
        throw exception::TableNotFoundException(expression->table_name);
    }

    auto *table = database.table(expression->table_name);
    if (table->schema().contains(expression->column_name) == false)
    {
        throw exception::ColumnNotFoundException(expression->table_name, expression->column_name);
    }

    const auto column_index = table->schema().column_index(expression->column_name);
    assert(column_index.has_value());
    const auto &column = table->schema()[column_index.value()];
    const auto index_exists = column.has_index(expression->index_name);
    if (expression->if_not_exists == false && index_exists)
    {
        throw exception::IndexAlreadyExistsException(expression->table_name, expression->column_name);
    }
    else if (index_exists)
    {
        return plan;
    }

    const auto id = plan.get_new_operator_ID("CREATE_INDEX");
    expression::Attribute attribute{std::move(expression->column_name), {std::move(expression->table_name)}};
    plan[id] = std::make_unique<expression::CreateIndexOperator>(
        std::move(attribute), std::move(expression->index_name), expression->is_unique, expression->type);

    return plan;
}

const Plan Builder::buildInsertPlan(Database &database, const parser::QueryParser &parser)
{
    Plan plan;

    const auto insert_operator_name = plan.get_new_operator_ID("INSERT");
    const auto insert_expression = parser.extractINSERT();

    if (insert_expression.has_value() == false)
    {
        throw exception::CanNotInsertException();
    }

    const auto table_exists = database.table_exists(insert_expression->table_name);
    if (table_exists == false)
    {
        throw exception::TableNotFoundException(insert_expression->table_name);
    }

    auto table = database.table(insert_expression->table_name);
    expression::Attributes attributes;
    for (const auto &column_name : insert_expression->column_names)
    {
        attributes.push_back({column_name, table->name()});
    }

    for (auto i = 0u; i < table->schema().size(); i++)
    {
        const auto &column = table->schema().column(i);
        if (column.is_nullable() == false)
        {
            const auto &table_attribute = table->schema().attribute(i);
            auto iterator = std::find_if(attributes.begin(), attributes.end(),
                                         [&table_attribute](const expression::Attribute &attribute) {
                                             return attribute.name == table_attribute.name;
                                         });
            if (iterator == attributes.end())
            {
                throw exception::ColumnCanNotBeNull(table->name(), table_attribute.name);
            }
        }
    }

    plan[insert_operator_name] = std::make_unique<expression::InsertOperator>(insert_expression->table_name, attributes,
                                                                              insert_expression->values_rows);

    return plan;
}

const Plan Builder::buildUpdatePlan(Database &database, const parser::QueryParser &parser)
{
    Plan plan;

    const auto update_expression = parser.extractUPDATE();

    AttributeToTableNameMapping attribute_mapping;
    TableAliasMapping table_references;
    std::vector<const table::Table *> source_tables;

    if (update_expression.from.has_value())
    {
        for (parser::TableDescr &source_desc : update_expression.from.value().tables())
        {
            const auto &source = source_desc.first;
            const auto &alias = source_desc.second;

            if (!database.table_exists(source))
            {
                throw exception::TableNotFoundException(source);
            }

            // check for proper disambiguation (via alias) in case of self-joins:
            const auto &table_name_iter = table_references.find(alias.value_or(source));

            if (table_name_iter != table_references.end())
            { // if name already exists...
                const auto is_alias_not_unique = table_references.find(alias.value()) != table_references.end();
                if (!alias || is_alias_not_unique)
                {
                    // not specifying an alias is an error, in this case
                    throw exception::MultipleTableReferences(source);
                }
            }

            // a table can be referenced with a user specified alias or its name
            table_references[alias.value_or(source)] = {database[source], static_cast<bool>(alias)};
            source_tables.push_back(database[source]); // for convenient, ordered table iteration
        }

        // check join predicates:
        expression::Attributes pred_attrs;
        for (const auto &predicate : update_expression.from.value().predicates())
        {
            auto attributes = expression::PredicateAnalyzer::attributes(predicate);
            std::move(attributes.begin(), attributes.end(), std::back_inserter(pred_attrs));
        }

        if (update_expression.where.has_value())
        {
            auto attributes = expression::PredicateAnalyzer::attributes(update_expression.where.value());
            std::move(attributes.begin(), attributes.end(), std::back_inserter(pred_attrs));
        }

        for (const auto &pred_attr : pred_attrs)
        {
            populateAttributeToTableMappingAndCheck(attribute_mapping, table_references, pred_attr);
        }
    }

    // check projection attribute- and table references.
    // Note: final-result schema will be created in the last step.
    for (const auto &update : update_expression.updates)
    {
        populateAttributeToTableMappingAndCheck(attribute_mapping, table_references, update.first);
    }

    std::optional<expression::Predicate> predicate = std::nullopt;
    if (update_expression.where.has_value())
    {
        predicate.emplace(resolveTableReferences(attribute_mapping, update_expression.where.value()));
    }

    auto top_node =
        addFromNodes(database, plan, update_expression.from.value().from_description->descr, attribute_mapping);

    if (predicate.has_value())
    {
        top_node = addSelection(plan, predicate.value(), top_node);
    }

    const auto update_operator_name = plan.get_new_operator_ID("UPDATE");

    plan[update_operator_name] = std::make_unique<expression::UpdateOperator>(update_expression.updates);
    plan.connect(top_node, update_operator_name);

    return plan;
}

const Plan Builder::build(Database &database, const beedb::parser::QueryParser &parser)
{
    switch (parser.type())
    {
    case parser::QueryParser::QueryType::SELECT:
        return buildSelectPlan(database, parser);
    case parser::QueryParser::QueryType::INSERT:
        return buildInsertPlan(database, parser);
    case parser::QueryParser::QueryType::CREATE_TABLE:
        return buildCreateTablePlan(database, parser);
    case parser::QueryParser::QueryType::CREATE_INDEX:
        return buildCreateIndexPlan(database, parser);
    case parser::QueryParser::QueryType::UPDATE:
        return buildUpdatePlan(database, parser);
    default:
        throw exception::LogicalException(
            "Plan construction not (yet) implemented for this query type! (QueryParser::QueryType::" +
            std::to_string(static_cast<std::uint32_t>(parser.type())) + ")");
    }
}