/*------------------------------------------------------------------------------*
 * Architecture & Implementation of DBMS                                        *
 *------------------------------------------------------------------------------*
 * Copyright 2020 Databases and Information Systems Group TU Dortmund           *
 * Visit us at                                                                  *
 *             http://dbis.cs.tu-dortmund.de/cms/en/home/                       *
 *                                                                              *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS      *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,  *
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL      *
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR         *
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,        *
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR        *
 * OTHER DEALINGS IN THE SOFTWARE.                                              *
 *                                                                              *
 * Authors (alphabetical list):                                                 *
 *          Maximilian Berens   <maximilian.berens@tu-dortmund.de>              *
 *          Roland Kühn         <roland.kuehn@cs.tu-dortmund.de>                *
 *          Jan Mühlig          <jan.muehlig@tu-dortmund.de>                    *
 *------------------------------------------------------------------------------*
 */

#include "plan/logical/plan.h"
#include <algorithm>
#include <iterator>
#include <util/text_table.h>

using namespace beedb::plan::logical;
using namespace beedb::expression;

OutgoingSchema Plan::produced_attributes(const NodeIDType &nid) const
{

    // First, add attributes that are added by this operator
    Attributes outgoing_schema = _node_data.at(nid)->additional_attributes;

    // multiple operators also forwards all incoming attributes:
    if (_node_data.at(nid)->forwards_schema)
    {
        NodeIDs inc;
        try
        {
            inc = this->_incoming_node_ids.at(nid);
        }
        catch (const std::out_of_range &)
        {
            // this is fine, there might be none
        }

        // iterate over incoming edges and insert edge data into outgoing_schema:
        for (const auto &inc_node_id : inc)
        {
            const auto &provided_schema = _edge_data.at({inc_node_id, nid});
            for (const auto &attr : provided_schema)
            {
                outgoing_schema.push_back(attr);
            }
        }
    }

    return outgoing_schema;
}

bool Plan::requirements_satisfied(const NodeIDType nid) const
{
    // first, collect incoming attributes (stored at incoming edges)
    std::set<Attribute> incoming_schema;
    NodeIDs inc;
    try
    {
        inc = this->_incoming_node_ids.at(nid);
    }
    catch (const std::out_of_range &)
    {
        // this is fine; there might be none
    }

    // iterate over incoming edges and collect edge data/incoming schemas:
    for (const auto &incoming_node : inc)
    {
        // note: there might be multiple incoming edges for (multiway) join operators
        const auto &provided_schema = _edge_data.at({incoming_node, nid});
        incoming_schema.insert(provided_schema.begin(),
                               provided_schema.end()); // why not "merge", C++, why..
    }

    // then test if required attributes are an actual subset of incoming attributes
    Attributes intersection;
    //    std::set_intersection(incoming_schema.begin(), // given attributes
    //                          incoming_schema.end(),
    //                          _node_data.at(nid)->requirements.begin(), // desired attributes
    //                          _node_data.at(nid)->requirements.end(),
    //                          std::inserter(intersection,
    //                                        intersection.begin()));

    // requirement is satisfied, if intersection is the same as "requirements"
    // if requirements are fully satisfied, attributes will be missing in "intersection"
    //    return intersection == _node_data.at(nid)->requirements;
    return intersection.size() == _node_data.at(nid)->requirements.size();
}

void Plan::print_table() const
{
    std::function<void(util::TextTable &, const NodeIDType &, const std::uint32_t)> recursive_add;

    recursive_add = [this, &recursive_add](util::TextTable &table, const NodeIDType &node_id,
                                           const std::uint32_t intend) {
        std::string name = std::string(intend, ' ') + this->_node_data.at(node_id)->operator std::string();
        table.push_back({{std::move(name)}});
        const auto &children = this->incoming_nodes(node_id);
        for (const auto &child : children)
        {
            recursive_add(table, child, intend + 2);
        }
    };

    util::TextTable text_table;
    text_table.header({{"Operator"}});
    const auto &root = this->find_root();
    if (root.has_value())
    {
        recursive_add(text_table, root.value(), 0u);
    }
    std::cout << text_table << std::endl;
}