/*------------------------------------------------------------------------------*
 * Architecture & Implementation of DBMS                                        *
 *------------------------------------------------------------------------------*
 * Copyright 2020 Databases and Information Systems Group TU Dortmund           *
 * Visit us at                                                                  *
 *             http://dbis.cs.tu-dortmund.de/cms/en/home/                       *
 *                                                                              *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS      *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,  *
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL      *
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR         *
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,        *
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR        *
 * OTHER DEALINGS IN THE SOFTWARE.                                              *
 *                                                                              *
 * Authors (alphabetical list):                                                 *
 *          Maximilian Berens   <maximilian.berens@tu-dortmund.de>              *
 *          Roland Kühn         <roland.kuehn@cs.tu-dortmund.de>                *
 *          Jan Mühlig          <jan.muehlig@tu-dortmund.de>                    *
 *------------------------------------------------------------------------------*
 */

#include <table/table_disk_manager.h>

using namespace beedb::table;

TableDiskManager::TableDiskManager(disk::BufferManager &buffer_manager) : _buffer_manager(buffer_manager)
{
}

void TableDiskManager::add_row(Table &table, beedb::table::Tuple &&tuple)
{
    auto [page, _] = this->add_row(table, tuple);
    this->_buffer_manager.unpin(page, true);
}

Tuple TableDiskManager::add_row(Table &table, const Schema &schema, Tuple &&tuple)
{
    auto [page, offset] = this->add_row(table, tuple);
    return Tuple(schema, page->id(), offset, (*page)[offset]);
}

std::pair<beedb::disk::Page *, std::size_t> TableDiskManager::add_row(beedb::table::Table &table,
                                                                      beedb::table::Tuple &tuple)
{
    const auto page_id = this->find_page_for_row(table);
    auto page = this->_buffer_manager.pin(page_id);

    const auto offset = page->append(tuple.data(), table.schema().row_size());
    return {page, offset};
}

std::vector<beedb::table::Tuple> TableDiskManager::read_rows(disk::Page *page, const Schema &schema)
{
    std::vector<Tuple> rows;
    auto offset = 0u;
    while (offset < page->size())
    {
        table::Tuple row(schema, page->id(), offset, (*page)[offset]);
        rows.push_back(std::move(row));
        offset += schema.row_size();
    }

    return rows;
}

void TableDiskManager::update_row(const Tuple &tuple)
{
    assert(tuple.page_id() != disk::Page::INVALID_PAGE_ID);
    assert(tuple.in_page_offset() >= 0);

    auto page = this->_buffer_manager.pin(tuple.page_id());
    page->overwrite(tuple.in_page_offset(), tuple.data(), tuple.schema().row_size());

    this->_buffer_manager.unpin(tuple.page_id(), true);
}

beedb::disk::Page::page_id TableDiskManager::find_page_for_row(Table &table)
{
    auto starting_page_id = table.page_id();
    if (table.last_page_id() != disk::Page::INVALID_PAGE_ID)
    {
        starting_page_id = table.last_page_id();
    }

    auto page = this->_buffer_manager.pin(starting_page_id);

    const auto needed = table.schema().row_size();
    while (page)
    {
        if (page->free_space() > needed)
        {
            break;
        }
        if (page->has_next_page())
        {
            const auto next_page_id = page->next_page_id();
            this->_buffer_manager.unpin(page, false);
            page = this->_buffer_manager.pin(next_page_id);
        }
        else
        {
            auto new_page = this->_buffer_manager.allocate();
            page->next_page_id(new_page->id());
            this->_buffer_manager.unpin(page, true);
            table.last_page_id(new_page->id());
            page = new_page;
            break;
        }
    }

    const auto page_id = page->id();
    this->_buffer_manager.unpin(page, false);
    return page_id;
}